(* ML types for abstract syntax trees of expressions *)


(* Surface expressions (with a few phrase categories merged).
   Adapted from Figures 3 and 4 in Lingay definition (version 0.2). *)

structure Ssyn = struct

type var = string
type label = string
type linopt = bool
type tyname = string

type ty = Types.ty
datatype literal = Bool of bool | Int of string
datatype infixOp = IntPlus | IntMinus | IntLessThan
datatype classkind = Imp | Fun | Lin
datatype safetag = 
	 None | Safe | Writesafe | Argsafe | Metasafe | MetasafeLin

datatype 'a exp = 
    Oncefn of 'a * 'a match
  | Fn of 'a * 'a match
  | Rec of 'a * 'a match
  | Catchcont of 'a * 'a match
  | If of 'a * 'a exp * 'a exp * 'a exp
  | Assign of 'a * label * 'a exp
  | Eq of 'a * 'a exp * 'a exp
  | VS of 'a * 'a exp * 'a exp
  | Dollar of 'a * 'a exp * 'a exp
  | Infix of 'a * infixOp * 'a exp * 'a exp
  | Merge of 'a * 'a exp * 'a exp
  | Cast of 'a * 'a exp * ty
  | Tag of 'a * label * 'a exp
  | Tok of 'a * label
  | Dot of 'a * 'a exp * label
  | Prom of 'a * 'a exp
  | Der of 'a * 'a exp
  | Force of 'a * 'a exp
  | Fold of 'a * ty * 'a exp
  | Unfold of 'a * 'a exp
  | Juxta of 'a * 'a exp * 'a exp
  | New of 'a * 'a exp
  | Make of 'a * 'a exp * 'a exp
  | Ref of 'a * tyname * 'a exp
  | Deref of 'a * 'a exp
  | Var of 'a * var
  | Lit of 'a * literal
  | Unit of 'a
  | Query of 'a * label
  | Product of 'a * (label * 'a exp) list
  | Split of 'a * 'a exp * recordpat * 'a exp
  | Case of 'a * 'a exp * 'a caseclause list
  | Root of 'a * linopt
  | Extend of 'a * linopt * 'a exp * 'a exp * 'a exp
  | Class of 'a * classkind * ty * 'a extendclause * ty * 
             'a methodbody list * 'a match
  | Reftype of 'a * tyname * ty * 'a exp
  | DecExp of 'a * 'a dec * 'a exp

and 'a dec = 
    Val of pattern * 'a exp
  | RecVal of var * ty * 'a exp
  | Funn of var * (pattern * ty) list * 'a exp          (* newly added *)
  | RecFunn of var * (pattern * ty) list * ty * 'a exp  (* newly added *)
  | It of 'a exp

and 'a match = 
    VarMatch of var * ty * 'a exp
  | PatMatch of pattern * ty * 'a exp
  | FoldMatch of ty * pattern * 'a exp

and pattern = 
    VarPat of var
  | RecPat of recordpat
  | FoldPat of ty * pattern
  | WildPat

and recordpat = 
    GenRecPat of (label * pattern) list * var
  | RigidRecPat of (label * pattern) list
  | FlexRecPat of (label * pattern) list
  | VSRecPat of pattern * pattern

and 'a caseclause = 
    VarClause of label * var * 'a exp
  | PatClause of label * pattern * 'a exp
  | TokClause of label * 'a exp

and 'a extendclause = 
    NoExtend
  | StdExtend of 'a exp * ty
  | AbbrevExtend of var

and 'a methodbody = 
    Method of label * pattern * 'a exp

datatype 'a topdec =
    TypeDec of 'a * tyname * ty
  | DatatypeDec of 'a * tyname * (label * ty option) list
  | Dec of 'a * safetag * 'a dec
  | Empty

end (* structure Ssyn *) ;


(* Underlying expressions *)

structure Absyn = struct

local open Types in

(* ASTs for underlying expressions *)

datatype const_exp = Bool of bool | Int of int
type infixOp = Ssyn.infixOp

datatype 'a exp = Var of 'a * var
	     | Const of 'a * const_exp
	     | Product of 'a * (label * 'a exp) list
	     | Merge of 'a *  'a exp * 'a exp
	     | Split of 'a * 'a exp * (label * var) list * var * 'a exp
	     | Inject of 'a * label * 'a exp
	     | Match of 'a * 'a exp * 'a case_exp list
	     | Prom of 'a * 'a exp
	     | Der of 'a * 'a exp
	     | Force of 'a * 'a exp
	     | Fold of 'a * ty * 'a exp
	     | Unfold of 'a * 'a exp
	     | Fun of 'a * var * ty * 'a exp  (* really oncefn *)
	     | App of 'a * 'a exp * 'a exp
	     | Eq of 'a * 'a exp * 'a exp
             | Infix of 'a * infixOp * 'a exp * 'a exp
	     | If of 'a * 'a exp * 'a exp * 'a exp
	     | Rec of 'a * var * ty * 'a exp
	     | Catchcont of 'a * var * ty * 'a exp
	     | Catchcopy of 'a * var * ty * 'a exp  (* obsolete *)
	     | Cast of 'a * 'a exp * ty 
	     | Triv of 'a * linopt
	     | Extend of 'a * linopt * 'a exp * 'a exp * 'a exp
	     | Constr of 'a * 'a exp
             | Make of 'a * 'a exp * 'a exp
and 'a case_exp = MatchCase of 'a * label * var * 'a exp

fun exp_top (e : 'a exp) : 'a = 
    case e of Var(x,_) => x
	    | Const(x,_) => x
	    | Product(x,_)=>x
	    | Merge(x,_,_)=>x
	    | Split(x,_,_,_,_)=>x
	    | Inject(x,_,_) =>x
	    | Match(x,_,_)=>x
	    | Prom(x,_)=>x
	    | Der(x,_)=>x
	    | Force(x,_)=>x
	    | Fold(x,_,_)=>x
	    | Unfold(x,_)=>x
	    | Fun(x,_,_,_)=>x
	    | App(x,_,_)=>x
	    | Eq(x,_,_)=>x
            | Infix(x,_,_,_)=>x
	    | If(x,_,_,_)=>x
	    | Rec(x,_,_,_)=>x
	    | Catchcont(x,_,_,_)=>x
	    | Catchcopy(x,_,_,_)=>x
	    | Cast(x,_,_)=>x
	    | Triv(x,_)=>x
	    | Extend(x,_,_,_,_)=>x
	    | Constr(x,_)=>x
            | Make(x,_,_)=>x
fun case_top (MatchCase(x,_,_,_))= x

fun exp_map (f: 'a -> 'b) (e : 'a exp) : 'b exp = let val m = exp_map f in
    case e of Var(x,z) => Var(f x, z)
	    | Const(x,ce) => Const(f x, ce)
	    | Product(x,es) => Product(f x, map (fn (l,e) => (l, m e)) es)
	    | Merge(x,e1,e2)=> Merge(f x, m e1, m e2)
	    | Split(x,e1,lvs,zz,e2)=>Split(f x, m e1, lvs, zz, m e2)
	    | Inject(x,k,e) => Inject(f x, k, m e)
	    | Match(x,e,cases) => Match(f x,m e, case_map f cases)
	    | Prom(x,e)=>Prom(f x, m e)
	    | Der(x,e)=>Der(f x,m e)
	    | Force(x,e)=>Force(f x,m e)
	    | Fold(x,t,e)=>Fold(f x,t,m e)
	    | Unfold(x,e)=>Unfold(f x, m e)
	    | Fun(x,z,t,e)=>Fun(f x,z,t,m e)
	    | App(x,e1,e2)=>App(f x,m e1, m e2)
	    | Eq(x,e1,e2)=>Eq(f x,m e1, m e2)
            | Infix(x,i,e1,e2)=>Infix(f x,i,m e1, m e2)
	    | If(x,e1,e2,e3)=>If(f x, m e1, m e2, m e3)
	    | Rec(x,z,t,e)=>Rec(f x,z,t,m e)
	    | Catchcont(x,z,t,e)=>Catchcont(f x,z,t,m e)
	    | Catchcopy(x,z,t,e)=>Catchcopy(f x,z,t,m e)
	    | Cast(x,e,t)=>Cast(f x,m e, t)
	    | Triv(x,l)=>Triv(f x,l)
	    | Extend(x,l,e1,e2,e3)=>Extend(f x,l,m e1,m e2,m e3)
	    | Constr(x,e)=>Constr(f x, m e)
            | Make(x,e1,e2)=>Make(f x, m e1, m e2)
    end
and case_map f cases = map (fn (MatchCase(x,k,z,e)) => MatchCase(f x,k,z,exp_map f e)) cases

fun get_labelled_vars  (e : 'a exp) : (string * 'a) list = 
    let val g = get_labelled_vars 
        fun without (z : string, l) = List.filter (fn (w,_) => w <> z) l
        fun withoutL (zl, l) = foldl without l zl
        fun case_vars (MatchCase(x,k,z,e)) = without (z, g e)
    in
    case e of Var(x,z) => [(z,x)]
	    | Const(x,ce) => []
	    | Product(x,es) => List.concat (map (fn (l,e) => g e) es)
	    | Merge(x,e1,e2)=> (g e1)@(g e2)
	    | Split(x,e1,lvs,zz,e2)=> withoutL (zz::(map (fn (l,x)=> x) lvs), (g e1) @ (g e2))
	    | Inject(x,k,e) => g e
	    | Match(x,e,cases) => (g e)@ (List.concat (map case_vars cases))
	    | Prom(x,e)=> g e
	    | Der(x,e)=> g e
	    | Force(x,e)=> g e
	    | Fold(x,t,e)=> g e
	    | Unfold(x,e)=> g e
	    | Fun(x,z,t,e)=> without (z, g e)
	    | App(x,e1,e2)=> (g e1) @ (g e2)
	    | Eq(x,e1,e2)=> (g e1) @ (g e2)
	    | Infix(x,i,e1,e2)=> (g e1) @ (g e2)
	    | If(x,e1,e2,e3)=> (g e1) @ (g e2) @ (g e3)
	    | Rec(x,z,t,e)=> without(z, g e)
	    | Catchcont(x,z,t,e)=> without (z, g e)
	    | Catchcopy(x,z,t,e)=> without (z, g e)
	    | Cast(x,e,t)=> g e
	    | Triv(x,l)=> []
	    | Extend(x,l,e1,e2,e3)=>(g e1)@(g e2)@(g e3)
	    | Constr(x,e)=> g e
            | Make(x,e1,e2)=>(g e1)@(g e2)
    end

fun get_vars e = map (fn (z,_) => z) (get_labelled_vars e)

fun exptostring e = 
  let val ets = exptostring
      fun ces l = concatWith ", " (map (fn (k,e) => k ^ "= " ^ (exptostring e)) l)
      fun splitting l = concatWith ", " (map (fn (k,x) => k ^ "=" ^ x) l)
      fun cases l = concatWith "| " (map (fn MatchCase (_,k,x,e) => k^" "^x^" => "^(exptostring e)) l)
      fun lin true = "linear " | lin false = ""
  in
  case e of Var(_, x) => x
      | Const(_,c) => (case c of Bool b => Bool.toString b
			       | Int i => Int.toString i)
      | Product (_,es) => "{"^(ces es)^"}"
      | Merge (_,e1,e2) => (ets e1)^ " +> "^(ets e2)
      | Split (_,e,lvs,rest,e') => "split "^(ets e)^" as " ^ (splitting lvs) ^ " +> "^rest^" in " ^ (ets e') ^ " end"
      | Inject(_,l,e) => "tag " ^l ^" "^(ets e)
      | Match (_,e,caselist) => "case "^(ets e)^" of "^(cases caselist)^" end"
      | Prom (_,e) => "prom "^(ets e)
      | Der (_,e) => "der "^(ets e)
      | Force (_,e) => "force "^(ets e)
      | Fold (_,t,e) => "fold "^(tytostring t)^" "^(ets e)
      | Unfold (_,e) => "unfold "^(ets e)
      | Fun(_,x,t,e) => "fn "^x^" : "^(tytostring t) ^ " => "^ (ets e)
      | App(_,e1,e2) => (ets e1) ^" $ "^ (ets e2)
      | Eq(_,e1,e2) => (ets e1) ^" = "^ (ets e2)
      | Infix(_,i,e1,e2) => (ets e1) ^
			    (case i of
			         Ssyn.IntPlus => " + "
			       | Ssyn.IntMinus => " - "
			       | Ssyn.IntLessThan => " < ") ^
                            (ets e2)
      | If(_,e1,e2,e3) => "if "^(ets e1) ^" then "^ (ets e2)^" else "^(ets e3)
      | Rec(_,x,t,e) => "rec "^x^" : "^(tytostring t) ^ " => "^ (ets e)
      | Catchcont(_,x,t,e) => "catchcont "^x^" : "^(tytostring t) ^ " => "^ (ets e)
      | Catchcopy(_,x,t,e) => "catchcopy "^x^" : "^(tytostring t) ^ " => "^ (ets e)
      | Cast(_,e,t) => (ets e)^ ":<" ^(tytostring t)
      | Triv(_, linopt) => (lin linopt)^ "root"
      | Extend(_,linopt,e1,e2,e3) => (lin linopt)^ "extend "^(ets e1)^" with "^(ets e2)^", "^(ets e3)^" end"
      | Constr (_,e) => "new "^(ets e)
      | Make (_,e1,e2) => "make "^(ets e1)^" "^(ets e2)
  end

fun show_exps es = concatWith "\n" (map exptostring es)
and show_exps_nl es = "\n" ^ (show_exps es) ^ "\n"

type ty_safe = Ssyn.safetag
val None = Ssyn.None
and Safe = Ssyn.Safe
and Writesafe = Ssyn.Writesafe
and Argsafe = Ssyn.Argsafe
and Metasafe = Ssyn.Metasafe
and MetasafeLin = Ssyn.MetasafeLin

datatype 'a decl = Exp of 'a exp
	      | Val of var * 'a exp * ty_safe
	      | Type of var * ty

(* For debugging *)

val displayTrans = ref false
fun displayExp e = 
    if !displayTrans then print ("TRANS: " ^ exptostring e ^ "\n") else ()
fun display (Exp e) = displayExp e
  | display (Val (v,e,_)) = displayExp e
  | display (Type _) = ()

(* Translating surface to underlying expressions 
   (resp. match and case clauses) *)

val map' = Pervasives.map'

exception Unfinished of string

fun dummyVar l = "_"^l

fun destVar (Ssyn.Var(_,v)) = SOME v 
  | destVar _ = NONE

fun formVals a [] e = e
  | formVals a ((l,pt)::rest) e = 
    Ssyn.DecExp (a, Ssyn.Val (pt, Ssyn.Var(a,dummyVar l)), formVals a rest e)

fun formFns a [] e = e
  | formFns a ((pat,t)::rest) e =
    Ssyn.Fn (a, Ssyn.PatMatch (pat, t, formFns a rest e))

fun lo Ssyn.Lin = true | lo _ = false

exception SuperLinopt of Ssyn.classkind  
    (* means parent class has wrong linearity status *)
fun checklo l ck = 
    if lo ck = l then () else raise SuperLinopt ck 
    (* should catch and report *)

val unitTy = Types.ProdTy([],[])
fun rootTy lo = Types.ClassImplTy (lo,unitTy,unitTy,unitTy)

(* Imperative hacks to retrieve types of named superclasses,
   and to make type environment available to translation: *)

val currTypeEnvStore = ref ([]: (var*Types.ty) list)
val currStaticEnvStore = ref ([]: (var*Types.ty) list)

exception UnknownSuperclass of var
fun lookupType v = 
    case List.find (fn (x,_) => x=v) (!currStaticEnvStore) of
         SOME (x,t) => t
       | NONE => raise UnknownSuperclass v ;

fun extendStuff a ck (Ssyn.StdExtend(e,t)) = (e,t)
  | extendStuff a ck Ssyn.NoExtend = (Ssyn.Root(a,lo ck), rootTy(lo ck))
  | extendStuff a ck (Ssyn.AbbrevExtend v) = (Ssyn.Var(a,v), lookupType v)

fun translateMeth Ssyn.Imp a pi_f' pi_m pi_m' (Ssyn.Method (m,pt,e)) =
    let val supercomps = compsOf pi_m
        val selfcomps = compsOf pi_m'
    in (m, Ssyn.Fn (a, Ssyn.VarMatch ("rw", RWall pi_f',
              Ssyn.Oncefn (a, Ssyn.PatMatch (pt, argtype pi_m' m, (*switched*)
            Ssyn.DecExp (a, Ssyn.Val 
		 (Ssyn.VarPat "super",
                     Ssyn.Product (a, map' supercomps (fn n => (n,
                           Ssyn.Fn (a, Ssyn.VarMatch ("x", argtype pi_m n,
                           Ssyn.Dollar (a, Ssyn.Juxta (a,
                             Ssyn.Dot (a, Ssyn.Var(a,"supervar"), n), 
                             Ssyn.Var (a,"rw")),
                             Ssyn.Var (a,"x")))))))),
            Ssyn.DecExp (a, Ssyn.Val 
		 (Ssyn.VarPat "self",
                     Ssyn.Product (a, map' selfcomps (fn n => (n,
                           Ssyn.Fn (a, Ssyn.VarMatch ("x", argtype pi_m' n,
                           Ssyn.Dollar (a, Ssyn.Juxta (a,
                             Ssyn.Dot (a, Ssyn.Var(a,"selfvar"), n), 
                             Ssyn.Var (a,"rw")),
                             Ssyn.Var (a,"x")))))))),
              e)))))))
    end
  | translateMeth _ _ _ _ _ _ = raise Unfinished "functional class sugar"

(* The main translation itself *)

(* Boring clauses: *)
fun translateExp (Ssyn.Var(a,v)) = Var(a,v)
  | translateExp (Ssyn.Lit(a,Ssyn.Bool b)) = Const(a,Bool b)
  | translateExp (Ssyn.Lit(a,Ssyn.Int i)) = 
        Const (a, Int (Option.valOf (Int.fromString i)))
  | translateExp (Ssyn.Product(a,complist)) =
        Product (a, map (fn (l,e) => (l,translateExp e)) complist)
  | translateExp (Ssyn.Merge(a,e1,e2)) =
        Merge (a, translateExp e1, translateExp e2)
  | translateExp (Ssyn.Tag(a,l,e)) = Inject (a,l,translateExp e)
  | translateExp (Ssyn.Case(a,e,ccl)) = 
        Match (a, translateExp e, map (translateCaseClause a) ccl)
  | translateExp (Ssyn.Prom(a,e)) = Prom (a,translateExp e)
  | translateExp (Ssyn.Der(a,e)) = Der (a,translateExp e)
  | translateExp (Ssyn.Force(a,e)) = Force (a,translateExp e)
  | translateExp (Ssyn.Fold(a,t,e)) = Fold (a,t,translateExp e)
  | translateExp (Ssyn.Unfold(a,e)) = Unfold (a,translateExp e)
  | translateExp (Ssyn.Oncefn(a,mt)) = 
        let val (v,t,e') = translateMatch a mt in Fun (a,v,t,e') end
  | translateExp (Ssyn.Dollar(a,e1,e2)) =
        App (a, translateExp e1, translateExp e2)
  | translateExp (Ssyn.Eq(a,e1,e2)) = 
        Eq (a, translateExp e1, translateExp e2)
  | translateExp (Ssyn.Infix(a,iop,e1,e2)) = 
        Infix (a, iop, translateExp e1, translateExp e2)
  | translateExp (Ssyn.If(a,e0,e1,e2)) =
        If (a, translateExp e0, translateExp e1, translateExp e2)
  | translateExp (Ssyn.Rec(a,mt)) =
        let val (v,t,e') = translateMatch a mt in Rec (a,v,t,e') end
  | translateExp (Ssyn.Catchcont(a,mt)) =
        let val (v,t,e') = translateMatch a mt in Catchcont (a,v,t,e') end
  | translateExp (Ssyn.Cast(a,e,t)) = Cast (a,translateExp e,t)
  | translateExp (Ssyn.Root(a,lo)) = Triv (a,lo)
  | translateExp (Ssyn.Extend(a,lo,e0,e1,e2)) =
        Extend (a,lo, translateExp e0, translateExp e1, translateExp e2)
  | translateExp (Ssyn.New(a,e)) = Constr (a, translateExp e)
  | translateExp (Ssyn.Make(a,e1,e2)) = 
        Make (a, translateExp e1, translateExp e2)
  | translateExp (Ssyn.Ref _) = raise Unfinished "ref"
  | translateExp (Ssyn.Deref _) = raise Unfinished "deref"
  | translateExp (Ssyn.Reftype _) = raise Unfinished "reftype"

(* Desugaring of derived forms: *)
  | translateExp (Ssyn.Fn(a,mt)) =
        let val (v,t,e') = translateMatch a mt in Prom(a,Fun(a,v,t,e')) end
  | translateExp (Ssyn.Assign(a,k,e)) =
        translateExp (Ssyn.Juxta (a,
            Ssyn.Dot (a, Ssyn.Dot (a,Ssyn.Var(a,"rw"),k), "write"), e))
  | translateExp (Ssyn.VS(a,e1,e2)) =
        Product (a, [("value",translateExp e1),("state",translateExp e2)])
        (* add to definition! *)
  | translateExp (Ssyn.Tok(a,k)) = Inject (a,k,Product(a,[]))
  | translateExp (Ssyn.Dot(a,e,k)) = 
        Split (a, translateExp e, [(k,"__x")], "__rest", Var(a,"__x"))
  | translateExp (Ssyn.Juxta(a,e1,e2)) =
        App (a, Der (a,translateExp e1), translateExp e2)
  | translateExp (Ssyn.Unit a) = Product(a,[])
  | translateExp (Ssyn.Query(a,k)) =
        translateExp (Ssyn.Juxta (a,
            Ssyn.Dot (a, Ssyn.Dot (a,Ssyn.Var(a,"rw"),k), "read"), 
                      Ssyn.Product(a,[])))
  | translateExp (Ssyn.Split(a,e0,Ssyn.GenRecPat(patlist,v),e1)) =
        Split (a, translateExp e0, map' patlist (fn (l,p)=>(l,dummyVar l)),
               v, translateExp (formVals a patlist e1))
        (* N.B. bug in definition here! *)
  | translateExp (Ssyn.Split(a,e0,Ssyn.FlexRecPat(patlist),e1)) =
        translateExp (Ssyn.Split(a,e0,Ssyn.GenRecPat(patlist,""),e1))
  | translateExp (Ssyn.Split(a,e0,Ssyn.RigidRecPat(patlist),e1)) =
        translateExp (Ssyn.Split(a,e0,Ssyn.GenRecPat(patlist,""),
            Ssyn.DecExp (a, Ssyn.Val (Ssyn.VarPat " ", 
                 Ssyn.Eq(a,Ssyn.Var(a,""), Ssyn.Product(a,[]))), e1)))
  | translateExp (Ssyn.Split(a,e0,Ssyn.VSRecPat(vpat,spat),e1)) =
        translateExp (Ssyn.Split(a,e0,Ssyn.GenRecPat
		 ([("value",vpat),("state",spat)],""), e1))

(* Val clauses - largely do duty for match clauses in definition: *)
  | translateExp (Ssyn.DecExp (a, Ssyn.Val (Ssyn.VarPat v, e0), e1)) =
        if destVar e0 = SOME v then translateExp e1  (* not strictly needed *)
        else Split (a, Product(a,[("k",translateExp e0)]),
                    [("k",v)], "", translateExp e1)
  | translateExp (Ssyn.DecExp (a, Ssyn.Val(Ssyn.WildPat, e0), e1)) =
        translateExp (Ssyn.DecExp (a, Ssyn.Val(Ssyn.VarPat "", e0), e1))
  | translateExp (Ssyn.DecExp (a, Ssyn.Val(Ssyn.FoldPat(ty,pat), e0), e1)) =
        translateExp (Ssyn.DecExp (a, Ssyn.Val(pat, Ssyn.Unfold(a,e0)), e1))
  | translateExp (Ssyn.DecExp (a, Ssyn.Val(Ssyn.RecPat rpat, e0), e1)) =
        translateExp (Ssyn.Split (a, e0, rpat, e1))
    (* for all non-Val declarations: *)
  | translateExp (Ssyn.DecExp (a, dec, e)) =
        translateExp (Ssyn.DecExp (a, desugarDec a dec, e))

(* Sugared class expressions: *)
  | translateExp (Ssyn.Class (a,ck,mty',exclause,fty,methbods,constrmatch)) = 
        let val (parent,parentType) = extendStuff a ck exclause
            val tenv = !currTypeEnvStore
            val (lo,pi_f,pi_m,tau_k) = destClassImplTy tenv parentType
            val _ = checklo lo ck
            val pi_m' = expandToProdTy tenv mty'
            val _ = checkExtends tenv pi_m' pi_m
            val pi_f' = mergeTypes tenv pi_f fty
            val tau_super = naturalTy lo pi_m pi_f'
            val tau_self = naturalTy lo pi_m' pi_f'
            val tau_k_pi_f = BangTy (FunTy (tau_k,pi_f))
        in translateExp
	      (Ssyn.Extend (a,lo,parent,
                 Ssyn.Fn (a, Ssyn.VarMatch ("supervar", tau_super,
                  Ssyn.Fn (a, Ssyn.VarMatch ("selfvar", tau_self,
                   Ssyn.Merge (a, Ssyn.Var (a,"supervar"),
                    Ssyn.Product (a, 
		     map (translateMeth ck a pi_f' pi_m pi_m') methbods)))))),
		 Ssyn.Fn (a, Ssyn.VarMatch ("super", tau_k_pi_f,
		  Ssyn.Fn (a, constrmatch)))))
        end

(* Auxiliary functions *)

and translateMatch a (Ssyn.VarMatch (v,t,e)) = (v,t,translateExp e)
  | translateMatch a (Ssyn.PatMatch (pat,t,e)) = 
        ("_d", t, translateExp 
		(Ssyn.DecExp (a, Ssyn.Val(pat, Ssyn.Var(a,"_d")), e)))
  | translateMatch a (Ssyn.FoldMatch (t,pat,e)) =
        translateMatch a (Ssyn.PatMatch (Ssyn.FoldPat(t,pat),t,e))

and translateCaseClause a (Ssyn.VarClause (k,v,e)) = 
        MatchCase (a,k,v,translateExp e)
  | translateCaseClause a (Ssyn.PatClause (k,pat,e)) = 
        translateCaseClause a (Ssyn.VarClause (k,"_d",
		(Ssyn.DecExp (a, Ssyn.Val(pat, Ssyn.Var(a,"_d")), e))))
  | translateCaseClause a (Ssyn.TokClause (k,e)) =
        translateCaseClause a (Ssyn.PatClause 
                (k, Ssyn.RecPat (Ssyn.RigidRecPat []), e))

and desugarDec a (Ssyn.Val stuff) = Ssyn.Val stuff
  | desugarDec a (Ssyn.RecVal (v,t,e)) = 
         Ssyn.Val (Ssyn.VarPat v, Ssyn.Rec (a, Ssyn.VarMatch (v,t,e)))
  | desugarDec a (Ssyn.Funn (v,patlist,e)) =
         Ssyn.Val (Ssyn.VarPat v, formFns a patlist e)
  | desugarDec a (Ssyn.RecFunn (v,patlist,t,e)) =
         Ssyn.Val (Ssyn.VarPat v, 
                   Ssyn.Rec (a, Ssyn.VarMatch (v,t,formFns a patlist e)))
  | desugarDec a (Ssyn.It e) = Ssyn.Val (Ssyn.VarPat "it", e)


(* Translating top-level declarations *)

local
fun patFns a [] exp = exp
  | patFns a ((pat,ty)::rest) exp = 
     Ssyn.Fn (a, Ssyn.PatMatch (pat, ty, patFns a rest exp))
fun multiArrowType [] ty = ty
  | multiArrowType ((_,ty0)::rest) ty =
    BangTy (FunTy (ty0, multiArrowType rest ty))
in
fun translateTopdec (Ssyn.Empty) = []
  | translateTopdec (Ssyn.TypeDec(a,t,ty)) = [Type(t,ty)]
  | translateTopdec (Ssyn.DatatypeDec(a,t,clist)) =
               let val t' = "_"^t
                   fun rename_t (l, SOME u) = (l, tySubst (t,NamedTy t') u)
                     | rename_t (l, NONE) = (l, ProdTy([],[]))
	       in
               Type (t, RecTy(t', SumTy (map rename_t clist))) ::
                   (* alpha conversion seems to do the trick here *)
               map' clist 
		   (fn (k,SOME ty) => Val (k, 
			 Prom(a, Fun(a,"x",ty, Fold(a, Types.NamedTy t, 
			   Inject(a,k, Var(a,"x"))))),
                         Ssyn.None)
		     | (k,NONE) => Val (k,
                         Fold(a, Types.NamedTy t, 
			   Inject(a,k,Product(a,[]))),
			 Ssyn.None))
               end
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.Val(Ssyn.VarPat v,exp))) =
              [Val (v,translateExp exp,st)]
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.Val(Ssyn.WildPat,exp))) =
              [Val ("",translateExp exp,st)]
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.Val(Ssyn.FoldPat(ty,pt),exp))) =
     translateTopdec (Ssyn.Dec(a,st,Ssyn.Val(pt, Ssyn.Unfold(a,exp))))
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.Val(Ssyn.RecPat rp,exp))) = 
              raise Unfinished "top-level val with pattern"
              (* - will require substantial changes to Lingay.sml
                 and Semantics.sml. Also, definition needs fixing! *)
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.RecVal(v,ty,exp))) =
              [Val (v,Rec(a,v,ty,translateExp exp),st)]
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.It exp)) =
              [Val ("it",translateExp exp,st)]
    (* newly added "fun" and "recfun" forms: *)
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.Funn(v,plist,exp))) =
     translateTopdec (Ssyn.Dec (a,st,Ssyn.Val (Ssyn.VarPat v,
					 patFns a plist exp)))
  | translateTopdec (Ssyn.Dec(a,st,Ssyn.RecFunn(v,plist,ty,exp))) =
     translateTopdec (Ssyn.Dec (a,st,Ssyn.Val (Ssyn.VarPat v,
		 Ssyn.Rec (a, Ssyn.VarMatch (v, multiArrowType plist ty,
                   patFns a plist exp)))))
end

end
end;
