(* File:    CBV_Games.sml
   Author:  John Longley
   Date:    July 2008 (started)
*)

(* SML implementation of the Abramsky/McCusker "Fam(C) construction",
   lifting an implementation of a CBN game model to a CBV one. *)


signature CBV_GAMES = sig

   eqtype ground = Pervasives.groundv
   type label = string
   type morphism
   type vmorph 
   val `        : vmorph * ground -> ground * morphism
   val evaluate : label -> vmorph -> ground * vmorph

   val INCLUDE  : ground * morphism -> vmorph  
   val GROUND   : (ground -> ground) -> vmorph
   val RELABEL  : (ground -> ground) -> vmorph
   val COND     : label * label * label -> vmorph
   val ID       : vmorph
   val THEN     : vmorph * vmorph -> vmorph
   val PRODUCT  : (label * vmorph) list -> vmorph
   val PROJ     : (label list * label list) list -> vmorph
   val PROJ'    : (label list * label list) list -> ground -> vmorph
   val PROD_INJ : label -> label list -> vmorph
   val CURRY    : label * label -> vmorph -> vmorph
   val UNCURRY  : label * label -> vmorph -> vmorph
   val BANG     : vmorph -> vmorph
   val DERELICT : vmorph
   val COMULT   : vmorph
   val CONTRACT : label list -> vmorph
   val LIFT     : vmorph -> vmorph 
   val UNIT     : vmorph
   val FLATTEN  : vmorph
   val STRENGTH : label -> vmorph
   val LPROMOTE : vmorph -> vmorph
   val FIX      : ground -> vmorph 
   val TAG      : label -> vmorph
   val CASES    : (label * vmorph) list -> vmorph
   val DISTRIB  : label -> label -> vmorph
   val THREAD   : label list -> vmorph
   val FORCE    : vmorph
   val CATCHCONT: vmorph

   val trace    : string -> vmorph -> vmorph
end ;


functor CBV_Games (
    structure CBN : CBN_GAMES where type ground = Pervasives.groundv
    structure CBN_Thread  : CBN_THREAD
    structure CBN_Control : CBN_CONTROL
    sharing type CBN.morphism = CBN_Thread.morphism = CBN_Control.morphism
    ) : CBV_GAMES = struct

open Pervasives

type ground = CBN.ground
type morphism = CBN.morphism
type vmorph = ground -> ground * morphism
infix ` %
fun F ` x = F x ;

fun trace tr F = fn x =>
   (TextIO.print ("    "^tr^" I:  "^ Pervasives.toString x^"\n") ;
    case F x of (y,s) =>
	(TextIO.print ("    "^tr^" J:  "^ Pervasives.toString y^"\n") ;
         (y, CBN.trace tr s)))

(* Crude evaluator *)

exception EvaluateError ;

local open CBN 
   (* val initVal0 = prodv [("ord",prodv []),("lin",prodv [])] *)
      fun flatRelab lab =
          {relabO = fn om as (ProdO (l,om')) =>
		       if l=lab then ProdO (l, MainO om') else om
		     | _ => raise EvaluateError ,
           relabP = fn pm as (ProdP (l,pm')) =>
		       if l=lab then
			   case pm' of MainP pm'' => ProdP (l,pm'')
				     | _ => raise EvaluateError
		       else pm
		     | _ => raise EvaluateError}
      fun addGround lab gv (prodv comps) =
	  prodv (Pervasives.updateAlist (lab,gv) comps)
	| addGround lab gv _ = raise EvaluateError
in
fun evaluate lab F = 
    case F ` blankv of (groundvals, f) =>
       case f % (Target (ProdO (lab, (BasicO blank)))) of 
	   (Target (ProdP (lab', (BasicP gv))), f') => 
              if lab'=lab then
		  let val groundvals' = addGround lab gv groundvals
                      val f'' = CBN.THEN (f', CBN.RELABEL (flatRelab lab))
		  in  (gv, fn _ => (groundvals',f''))
		  end
              else raise EvaluateError
	 | _ => raise EvaluateError 
end ;

fun INCLUDE (x,f) = fn _ => (x,f) ; 

fun GROUND h = fn x => (h x, CBN.ZERO) ;

fun RELABEL h = fn x => (h x, CBN.ID) ;

val ID = fn x => (x, CBN.ID) ;

fun op THEN (F,G) = 
    fn x => case F x of (y,f) => 
        case G y of (z,g) => (z, CBN.THEN (f,g)) ;

exception CompError of label
fun comp l (prodv xlist) = 
    errorLookup (fn l' => (print ("CompError "^l'^
                                  String.concat (map fst xlist)^"\n") ; 
			   CompError l'))
		xlist l
  | comp l y = (print (Pervasives.toString y) ; raise CompError l) ;

fun PRODUCT complist = fn x =>
    let val complist' = map' complist (fn (l,F) => (l,F (comp l x))) 
    in (prodv (map' complist' (fn (l,(x,f)) => (l,x))),
        CBN.PRODUCT (map' complist' (fn (l,(x,f)) => (l,f))))
    end ;

(* Generalized projection: a bit of work required for ground type moves *)

fun optrev (SOME l) = SOME (List.rev l)
  | optrev NONE = NONE

fun filterOpt [] = []
  | filterOpt (NONE::rest) = filterOpt rest
  | filterOpt ((SOME x)::rest) = x::filterOpt rest

fun prefixRest [] path = SOME path
  | prefixRest _ [] = NONE
  | prefixRest (x::xs) (y::ys) =
    if x=y then prefixRest xs ys else NONE

fun renamePath [] path = NONE
  | renamePath ((p,p')::rest) path =
    case prefixRest (List.rev p) path of
	SOME p'' => SOME (List.rev p' @ p'')  (* path = rev p @ p'' *) 
      | NONE => renamePath rest path
   (* backwards representation of paths in renaming is confusing! *)

fun applyRenaming ren data = 
   filterOpt (map' data (fn (labs,x) => 
	(case renamePath ren labs of
	     NONE => NONE
	   | SOME labs' => SOME (labs', x))))

fun separateComps (prodv []) = [([],prodv[])]
  | separateComps (prodv xlist) =
    List.concat (map' xlist (fn (l,x) => 
	map' (separateComps x) (fn (labs,y) => (l::labs,y))))
  | separateComps x = [([],x)] ;

exception CombineError

fun combineComps [([],x)] = x
(*| combineComps [] = raise CombineError *)
  | combineComps data = prodv (combineComps' data)
and combineComps' [] = []
  | combineComps' (all as ((l::labs,x)::rest)) =
    let val (lcomps, others) = 
	    List.partition
		(fn (l'::_,_) => l'=l | _ => false) all
        val lcomps' = map' lcomps (fn (labs,x) => (List.tl labs,x))
    in (l, combineComps lcomps') :: combineComps' others
    end 
  | combineComps' _ = [] ;

fun PROJ ren = fn x =>
   (combineComps (applyRenaming ren (separateComps x)),
    CBN.PROJ ren)

(* Variant to coerce result to a certain skeletal form *)

exception MatrixError

fun addToMatrix (prodv clist) (prodv clist') =
    prodv (mergeCompLists clist clist')
  | addToMatrix _ _ = raise MatrixError
and mergeCompLists [] clist' = clist'
  | mergeCompLists ((l,x)::clist) clist' =
    case List.partition (fn (l',y) => l'=l) clist' of
        ([], others) => (l,x) :: mergeCompLists clist others
      | ([(l',y)], others) => (l,addToMatrix x y) :: 
			      mergeCompLists clist others
      | _ => raise MatrixError ;

fun PROJ' ren matrix = fn x =>
   (addToMatrix matrix (combineComps (applyRenaming ren (separateComps x))),
    CBN.PROJ ren)

fun PROD_INJ lab labs = fn x =>
   (prodv ((lab,x) :: map' labs (fn l => (l,CBN.blank))),
    CBN.PROJ [([],[lab])])

exception CondError
fun COND (l1,l2,l3) = fn x =>
    case comp l1 x of
	boolv true  => (comp l2 x, CBN.PROJ [([l2],[])])
      | boolv false => (comp l3 x, CBN.PROJ [([l3],[])])
      | _ => raise CondError ;


(* Curry/uncurry: assume codomain is pointed *)

fun CURRY (l1,l2) F = fn x =>
   (CBN.blank, CBN.ABSORB (fn y => 
       case F (prodv [(l1,x),(l2,y)]) of (_,f) =>
           CBN.CURRY (l1,l2) f)) ;

fun UNCURRY (l1,l2) F = fn xy =>
   let val (_,f) = F (comp l1 xy) in
       (CBN.blank, CBN.SUPPLY (comp l2 xy) (CBN.UNCURRY (l1,l2) f))
   end ;

(* Bang stuff *)

fun BANG F = fn x => case F x of (y,f) => (y, CBN.BANG f) ;

val DERELICT = fn x => (x, CBN.DERELICT) ;

val COMULT = fn x => (x, CBN.COMULT) ;

fun CONTRACT labs = fn x => 
    (prodv (map (fn l => (l,x)) labs), CBN.CONTRACT) ;

(* Lifting stuff *)

fun LIFT F = INCLUDE (CBN.blank, CBN.LIFT F) ;

val UNIT = fn x => (CBN.blank, CBN.UNIT x) ;

val FLATTEN = fn x => (x, CBN.FLATTEN) ;

exception GroundError ;
fun STRENGTH lab = 
    fn (prodv comps) => 
        (CBN.blank, CBN.STRENGTH lab 
		    (fn y => prodv (updateAlist (lab,y) comps))) 
     | _ => raise GroundError ;

fun FIX x0 = fn x => (x0, CBN.FIX) ;

exception LPromoteError ;
local open CBN 
      fun CBN_promote f = CBN.THEN (COMULT, BANG f) ;
      val unmainRelab = {
	  relabP = fn (MainP pm) => pm | _ => raise LPromoteError,
	  relabO = MainO
      } 
in
fun LPROMOTE F =
    fn x => case F`x of (_,f) =>
	case CBN.% (f, Target (BasicO blank)) of
            (Target (BasicP z), f') =>
            (blank,
             CBN.THEN
                 (CBN_promote (CBN.THEN (f', RELABEL unmainRelab)),
                  CBN.UNIT z))
          | _ => raise LPromoteError  (* attempt to promote non-value *)
end ;

(* Sums: seems we don't need Inj tagging in CBN games after all... *)

fun TAG lab = RELABEL (fn x => injv (lab,x)) ;

exception CasesError ;
fun CASES caselist = 
    fn (injv (lab,x)) =>
       (case Pervasives.optLookup caselist lab of
	    SOME F => F x
	  | NONE => raise CasesError)
     | _ => raise CasesError ;

exception DistribError ;

fun distr sumlab (injv (summand,y)) lab x = 
    injv (summand, prodv [(sumlab,y), (lab,x)])
  | distr _ _ _ _ = raise DistribError ;

fun DISTRIB sumlab lab = 
    RELABEL (fn prodv [(lab1,x1),(lab2,x2)] =>
	if lab1=sumlab andalso lab2=lab then distr sumlab x1 lab x2
        else if lab2=sumlab andalso lab1=lab then distr sumlab x2 lab x1
        else raise DistribError
	      | _ => raise DistribError) ;


fun THREAD methods = fn x => (CBN.blank, CBN_Thread.THREAD methods) ;

val FORCE = fn x => (CBN.blank, CBN_Control.FORCE) ;

val CATCHCONT = fn x => (CBN.blank, CBN_Control.CATCHCONT) ;

end (* functor CBV_Games *) ;



(* Testing for PROJ:

val renaming = [ (["1","1"], ["a"]),
                 (["2","1"], ["a","b"]),
                 (["2"],     ["b","b"]) ] ;
val G = PROJ renaming ;
G (prodv [("2",intv 5)]) ;
G (prodv [("2", intv 5), ("1", prodv [("1", intv 6), ("2", intv 7)])]) ;

*)
