
(* Strategies for the control operators "catchcont" and "force". *)

signature CBN_CONTROL = sig
   type morphism
   val FORCE : morphism
   val CATCHCONT : morphism
end ;


functor CBN_Control (
   structure CBN_Games : CBN_GAMES
   where type ground = Pervasives.groundv
   ) : CBN_CONTROL = struct

open Pervasives 
open CBN_Games

type index = int 

(* Implementation of FORCE *)

structure Force = struct

type XG = CBN_Games.ground 
type XO = CBN_Games.omove 
type XP = CBN_Games.pmove 

datatype forceOmove =
    P1 of XG * index
  | P2 of XP * index
  | O3 
  | O4 of XO * index

datatype forcePmove =
    O1 of index
  | O2 of XO * index
  | P3 of XG
  | P4 of XP * index

exception ForceError

fun forceMove O3 = O1 0
  | forceMove (P1 (xg,i)) = if i=1 then P3 xg else raise ForceError
  | forceMove (O4 (xo,i)) = O2 (xo,i+1)
  | forceMove (P2 (xp,i)) = if i>0 then P4 (xp,i-1) else raise ForceError

fun codePmove (O1 i)     = Source (TagO (i, RightO (BasicO blank))) 
				       (* or prodv [] or something? *)
  | codePmove (O2(xo,i)) = Source (TagO (i, RightO (MainO xo)))
  | codePmove (P3 xg)    = Target (BasicP xg)
  | codePmove (P4(xp,i)) = Target (MainP (TagP (i,xp)))

fun decodeOmove (Source (TagP (i, RightP (BasicP xg)))) = P1 (xg,i)
  | decodeOmove (Source (TagP (i, RightP (MainP xp)))) = P2 (xp,i)
  | decodeOmove (Target (BasicO blank)) = O3
  | decodeOmove (Target (MainO (TagO (i,xo)))) = O4 (xo,i)
  | decodeOmove _ = raise ForceError

fun forceStrat() = strategy (fn om =>
	(codePmove (forceMove (decodeOmove om)), forceStrat()))

end (* structure Force *) ;

val FORCE : morphism = Force.forceStrat() ;


(* Implementation of CATCHCONT *)

structure Catchcont = struct

type Z = CBN_Games.ground   (* zeta  *)
type V = CBN_Games.ground   (* tau_0 *)
type XG = CBN_Games.ground  (* xi    *)
type XO = CBN_Games.omove 
type XP = CBN_Games.pmove 
type YG = CBN_Games.ground  (* tau_1 *)
type YO = CBN_Games.omove 
type YP = CBN_Games.pmove 

datatype ('a,'b) RQ = result of 'a | query of 'b 

datatype catchcontOmove =
    O1 of Z * index
  | O2 of XO * index
  | P3 of V * YG
  | P4 of YP
  | O5 
  | P6 of XG * index
  | P7 of XP * index
  | O8 
  | O9 of YO
  | P10 of XP * index   (* bang missing in definition! *)
  | O11 of XG
  | P12 of XG * index
  | P13 of XP * index
  | O14
  | O15 of YO

datatype catchcontPmove =
    P1 of XG * index
  | P2 of XP * index
  | O3 
  | O4 of YO
  | P5 of (V,Z) RQ
  | O6 of Z * index
  | O7 of XO * index
  | P8 of YG
  | P9 of YP
  | O10 of XO * index   (* bang missing in definition! *)
  | P11
  | O12 of Z * index
  | O13 of XO * index
  | P14 of V * YG
  | P15 of YP

type admin = {
     nextLeftIndex: int, 
     leftXiAssoc: (int*int) list, 
     leftRightAssoc: (int*int) list
     } ;

val initAdmin: admin = {
    nextLeftIndex = 2, 
    leftXiAssoc = [(1,0)], 
    leftRightAssoc = []
    } ;

fun addRight i (adm:admin) = {
    nextLeftIndex = #nextLeftIndex adm + 1,
    leftRightAssoc = (#nextLeftIndex adm,i):: #leftRightAssoc adm,
    leftXiAssoc = #leftXiAssoc adm
    } ;

fun addXi i (adm:admin) = {
    nextLeftIndex = #nextLeftIndex adm + 1,
    leftXiAssoc = (#nextLeftIndex adm,i):: #leftXiAssoc adm,
    leftRightAssoc = #leftRightAssoc adm
    } ;

datatype index' = Xi of int | Right of int

exception CatchcontError

fun getAssoc i (adm:admin) =
    (case Pervasives.optLookup (#leftXiAssoc adm) i of
	 SOME j => Xi j | NONE =>
	 (case Pervasives.optLookup (#leftRightAssoc adm) i of
	      SOME j => Right j | NONE => raise CatchcontError)) 

datatype status = INIT | RES of YG | RES' | QUE of XG | QUE' of admin

fun catchcontMove O5 INIT = (O3, INIT)
  (* result case *)
  | catchcontMove (P3(v,yg)) INIT = (P5 (result v), RES yg)
  | catchcontMove O8 (RES yg) = (P8 yg, RES')  
  | catchcontMove (O9 yo) RES' = (O4 yo, RES')
  | catchcontMove (P4 yp) RES' = (P9 yp, RES')
  | catchcontMove (O1(z,0)) RES' = (O6(z,0), RES')
  | catchcontMove (P6(xg,i)) RES' = (P1(xg,i), RES')
  | catchcontMove (O2(xo,i)) RES' = (O7(xo,i), RES')
  | catchcontMove (P7(xp,i)) RES' = (P2(xp,i), RES')
  (* query case *)
  | catchcontMove (O1(z,0)) INIT = (P5 (query z), INIT)
  | catchcontMove (O11 xg) INIT = (P11, QUE xg)
  | catchcontMove O14 (QUE xg) = (P1(xg,1), QUE' initAdmin)
  | catchcontMove (P3(v,yg)) (QUE' adm) = (P14(v,yg), QUE' adm)
  | catchcontMove (O15 yo) (QUE' adm) = (O4 yo, QUE' adm)
  | catchcontMove (P4 yp) (QUE' adm) = (P15 yp, QUE' adm)
  (* now the moves involving bang indices *)
  | catchcontMove (O1(z,0)) (QUE' adm) = (O12(z,0), QUE' adm)
  | catchcontMove (P12(xg,i)) (QUE' adm) = 
                  (P1(xg,#nextLeftIndex adm), QUE' (addRight i adm))
  | catchcontMove (O2(xo,i)) (QUE' adm) =
                  (case getAssoc i adm of
		       Xi j => (O10(xo,j), QUE' adm)
		     | Right j => (O13(xo,j), QUE' adm))
  | catchcontMove (P10(xp,i)) (QUE' adm) =
                  (P2(xp,#nextLeftIndex adm), QUE' (addXi i adm))
  | catchcontMove (P13(xp,i)) (QUE' adm) =
                  (P2(xp,#nextLeftIndex adm), QUE' (addRight i adm))
  | catchcontMove _ _ = raise CatchcontError

(* Coding/decoding operations to follow *)

fun codePmove (P1(xg,i)) = Source (LeftO (TagP (i, RightP (BasicP xg))))
  | codePmove (P2(xp,i)) = Source (LeftO (TagP (i, RightP (MainP xp))))
  | codePmove O3      = Source (RightO (BasicO blank))
  | codePmove (O4 yo) = Source (RightO (MainO (ProdO ("residue",yo))))
  | codePmove (P5(result v)) = 
      Target (BasicP (injv ("result", prodv [("value",v),("more",blank)])))
  | codePmove (P5(query z))  = 
      Target (BasicP (injv ("query", prodv [("arg",z),("resume",blank)])))
  | codePmove (O6(z,i)) = 
      Target (MainP (ProdP ("more", LeftP (TagO (i, RightO (BasicO z))))))
  | codePmove (O7(xo,i)) =
      Target (MainP (ProdP ("more", LeftP (TagO (i, RightO (MainO xo))))))
  | codePmove (P8 yg) =
      Target (MainP (ProdP ("more", RightP (BasicP yg))))
  | codePmove (P9 yp) =
      Target (MainP (ProdP ("more", RightP (MainP yp))))
  | codePmove (O10(xo,i)) =
      Target (MainP (ProdP ("resume", LeftP (TagO (i,xo)))))
  | codePmove P11 =
      Target (MainP (ProdP ("resume", RightP (BasicP blank))))
  | codePmove (O12(z,i)) =
      Target (MainP (ProdP ("resume", RightP (MainP (LeftP 
	(TagO (i, RightO (BasicO z))))))))
  | codePmove (O13(xo,i)) =
      Target (MainP (ProdP ("resume", RightP (MainP (LeftP 
	(TagO (i, RightO (MainO xo))))))))
  | codePmove (P14(v,yg)) =
      Target (MainP (ProdP ("resume", RightP (MainP (RightP (BasicP
        (prodv [("value",v),("residue",yg)])))))))
  | codePmove (P15 yp) =
      Target (MainP (ProdP ("resume", RightP (MainP (RightP (MainP
        (ProdP ("residue",yp))))))))

fun decodeOmove (Source (LeftP (TagO (i, RightO (BasicO z)))))
      = O1(z,i)
  | decodeOmove (Source (LeftP (TagO (i, RightO (MainO xo)))))
      = O2(xo,i)
  | decodeOmove (Source (RightP (BasicP 
                 (prodv [("value",v),("residue",yg)]))))
      = P3(v,yg)
  | decodeOmove (Source (RightP (MainP (ProdP ("residue",yp)))))
      = P4 yp
  | decodeOmove (Target (BasicO blank))
      = O5
  | decodeOmove (Target (MainO (ProdO ("more", LeftO (TagP (i, 
		 RightP (BasicP xg)))))))
      = P6(xg,i)
  | decodeOmove (Target (MainO (ProdO ("more", LeftO (TagP (i,
	         RightP (MainP xp)))))))
      = P7(xp,i)
  | decodeOmove (Target (MainO (ProdO ("more", RightO (BasicO blank)))))
      = O8
  | decodeOmove (Target (MainO (ProdO ("more", RightO (MainO yo)))))
      = O9 yo
  | decodeOmove (Target (MainO (ProdO ("resume", LeftO (TagP (i,xp))))))
      = P10(xp,i)
  | decodeOmove (Target (MainO (ProdO ("resume", RightO (BasicO xg)))))
      = O11 xg
  | decodeOmove (Target (MainO (ProdO ("resume", RightO (MainO (LeftO
                 (TagP (i, RightP (BasicP xg)))))))))
      = P12 (xg,i)
  | decodeOmove (Target (MainO (ProdO ("resume", RightO (MainO (LeftO
                 (TagP (i, RightP (MainP xp)))))))))
      = P13 (xp,i)
  | decodeOmove (Target (MainO (ProdO ("resume", RightO (MainO (RightO
		 (BasicO blank)))))))
      = O14
  | decodeOmove (Target (MainO (ProdO ("resume", RightO (MainO (RightO
                 (MainO (ProdO ("residue",yo)))))))))
      = O15 yo
  | decodeOmove _ = raise CatchcontError

fun catchcontStrat status = strategy (fn om =>
    let val (pm,status') = catchcontMove (decodeOmove om) status
    in  (codePmove pm, catchcontStrat status')
    end) ;

(* Also, remember to tweak typechecker and theory code re extra bang *)

end (* structure Catchcont *) ;

val CATCHCONT = (Catchcont.catchcontStrat Catchcont.INIT) ;

end (* functor Control *) ;


(* For testing catchcont: 

type moreType   = (bool->int) -o (bool->int) ;
type resumeType = !int -o (bool->int) -o {value:int, residue:bool->int} ;
type resultType = {value:int, more:moreType} ;
type queryType  = {arg:bool,  resume:resumeType} ;
type catchcontType = [| result of resultType, query of queryType |] ;

val destResult  = fn c:catchcontType => case c of
                     result R => R 
		   | query Q => {value=0, more=rec more:moreType => more}
                  end ;

val destQuery   = fn c:catchcontType => case c of
                     query Q => Q
                   | result R => {arg=false, resume=rec resume:resumeType => resume}
                  end ;

val c = catchcont t:bool->int => {value = t true, residue = t} ;
destQuery c .resume ;
it $ prom 5 ;
it $ (fn b:bool => if b then 6 else 7) ;
val r = it.residue ;
r true ;
r false ;

val c' = catchcont t:bool->int => {value = 9, residue = t} ;
*)