(* File:    Pervasives.sml
   Author:  John Longley
   Date:    August 2008 
*)

(* Pervasives and general utilities for Lingay runtime *)


structure VarSet = ListSetFn (struct type ord_key = string 
                                     val compare = String.compare end)
(* Available to the CM, but apparently not in the standard basis. *)


structure Pervasives = struct

infix THEN THENL ;
infix % ;
infix ` `` ;

(* List utilities *)

fun optLookup [] x = NONE
  | optLookup ((x',y)::rest) x =
    if x'=x then SOME y else optLookup rest x ;

fun optLookup' [] y = NONE
  | optLookup' ((x,y')::rest) y =
    if y'=y then SOME x else optLookup' rest y ;

fun errorLookup exn alist x =
    (case optLookup alist x of
        SOME y => y
      | NONE => raise exn x) ;

fun fst (x,y) = x and snd (x,y) = y ;

fun updateAlist pair [] = [pair]
  | updateAlist pair (pair'::rest) =
    if fst pair' = fst pair then pair::rest
    else pair'::updateAlist pair rest ;

fun map' l f = map f l ;
fun concatmap f l = List.concat (map f l) ;
fun concatmap' l f = concatmap f l ;
fun mapsnd f = map (fn (x,y) => (x,f y)) ;
fun all' l P = List.all P l ;
fun exists' l P = List.exists P l ;

fun remove x [] = []
  | remove x (y::l) = if y=x then l else y::remove x l
fun diff l1 [] = l1
  | diff l1 (x::l2) = diff (remove x l1) l2 ;
fun member x l = List.exists (fn y => y=x) l ;
fun insert x l = if member x l then l else x::l

fun merge [] l = l
  | merge (x::l) l' = merge l (insert x l')

fun mergeAll [] = []
  | mergeAll [l] = l
  | mergeAll (l::rest) = merge l (mergeAll rest) ;

(* Ground type data carried by game moves *)

type label = string ;

datatype groundv =
	 blankv 
       | boolv of bool 
       | intv of int 
       | prodv of (label * groundv) list
       | injv of label * groundv
       | foldv of groundv ;

fun toString blankv = "-"
  | toString (boolv b) = Bool.toString b
  | toString (intv i) = Int.toString i
  | toString (prodv complist) = "{" ^ listToString complist ^ "}"
  | toString (injv (l,x)) = "(tag " ^ l ^ " " ^ toString x ^ ")"
  | toString (foldv x) = "(fold " ^ toString x ^ ")"
and listToString [] = ""
  | listToString [lx] = labToString lx
  | listToString (lx::lxs) = labToString lx ^ ", " ^ listToString lxs
and labToString (l,x) = l ^ "=" ^ toString x ;

(* Ground type operations *)

datatype infixop  = IntPlus | IntMinus | IntLess

local exception PairError 
      exception IntError 
      val lookup = errorLookup (fn _ => PairError)
      fun getInt (intv i) = i | getInt _ = raise IntError
      fun strip blankv = NONE
        | strip (boolv b) = SOME (boolv b)
        | strip (intv i) = SOME (intv i)
        | strip (prodv clist) =
	  (case List.filter (fn (l,x) => x <> NONE)
		    (map' clist (fn (l,x) => (l,strip x))) of
	       [] => NONE
             | clist' => SOME (prodv (map' clist' (fn (l,x) => (l,valOf x)))))
        | strip (injv (l,x)) =
	  (case strip x of NONE => NONE | SOME y => SOME (injv (l,y)))
	| strip (foldv x) =
	  (case strip x of NONE => NONE | SOME y => SOME (foldv y))
      fun groundEq (boolv b, boolv b') = b=b'
        | groundEq (intv i, intv i') = i=i'
	| groundEq (prodv clist, prodv clist') =
	  List.length clist = List.length clist' andalso
	  all' clist (fn (l,x) => 
		List.exists (fn (l',x') => l'=l andalso groundEq (x,x'))
			    clist')
        | groundEq (injv (l,x), injv (l',x')) =
	  l=l' andalso groundEq (x,x')
	| groundEq (foldv x, foldv x') = groundEq (x,x')
	| groundEq (_,_) = false
      fun groundEq' (SOME x, SOME y) = groundEq (x,y)
	| groundEq' (NONE, NONE) = true
	| groundEq' (_,_) = false
in
fun GroundEq (prodv comps) =
    boolv (groundEq' (strip (lookup comps "#1"), strip (lookup comps "#2"))) 
  | GroundEq _ = raise PairError ;
fun GroundInfix iop (prodv comps) =
    let val i1 = getInt (lookup comps "#1")
        val i2 = getInt (lookup comps "#2")
    in case iop of 
	   IntPlus  => intv (i1+i2)
	 | IntMinus => intv (i1-i2)
	 | IntLess  => boolv (i1<i2)
    end 
  | GroundInfix _ _ = raise PairError
end ;

end ;
