
(* File:    Thread.sml
   Author:  John Longley
   Date:    July 2008 
*)

(* Thread operation for Lingay runtime.
   Full version for non-linear classes, incorporating an
   "outermost bang" to allow backtracking without snapback.
   The thread strategy is implemented concretely in terms of moves. *)

(* Full version of thread, of type

   {m: !( {f: {read: !(Lift S_f), write: !(S_f -o Lift 1)}} 
                        1   2              3       4        
                        +   +              -       +        

          -o (A_m -o B_m) ) }
              5      6
              +      -

   -o   ! ( {f: !S} -o  Lift {m: !(A_m -o B_m)} )
                 7       8         9      10
                 -       +         -      +

   We can think of the B_m here as incorporating a lifting.
   However, there is no lift on (A_m -o B_m) as a whole,
   so the interpretation of the syntactic type "\tau_self" needs
   a bit of flattening to match the LHS of the above type.
*)

functor Thread (
   eqtype SG; eqtype SO; eqtype SP; 
   eqtype AO; eqtype AP; eqtype BO; eqtype BP
   (* eqtypes useful for testing, but not very necessary for compilation *)
   ) = struct

open Pervasives 

type label = string
type index = int
type objectId = unit ref
datatype topIndex = wrap of int
fun unwrap (wrap n) = n ;
fun inc (wrap n) = wrap (n+1) ;

type ('a,'b) alist = ('a * 'b) list ;
type SGall = (label,SG) alist ;

exception LabelMismatch of label * label ;
fun check (m:label) (m':label) = 
   if m=m' then m else raise LabelMismatch (m,m') ;

datatype threadOmove =
   O1 of index * label * index * label   (* inner index and label first *)
 | O2 of SO * index * label * index * label 
 | P3 of SP * index * label * index * label 
 | O4 of SG * index * label * index * label 
 | O5 of AO * index * label 
 | P6 of BP * index * label 
 | P7 of SP * index * label * topIndex
 | O8 of SGall * topIndex
 | P9 of AP * index * label * topIndex
 | O10 of BO * index * label * topIndex

datatype threadPmove =
   P1 of SG * index * label * index * label 
 | P2 of SP * index * label * index * label 
 | O3 of SO * index * label * index * label 
 | P4 of index * label * index * label 
 | P5 of AP * index * label 
 | O6 of BO * index * label 
 | O7 of SO * index * label * topIndex
 | P8 of topIndex
 | O9 of AO * index * label * topIndex
 | P10 of BP * index * label * topIndex

fun outerAddrO (O1(_,_,j,m))   = SOME (j,m)
  | outerAddrO (O2(_,_,_,j,m)) = SOME (j,m)
  | outerAddrO (P3(_,_,_,j,m)) = SOME (j,m)
  | outerAddrO (O4(_,_,_,j,m)) = SOME (j,m)
  | outerAddrO (O5(_,j,m))   = SOME (j,m)
  | outerAddrO (P6(_,j,m))   = SOME (j,m)
  | outerAddrO _ = NONE ;

fun outerAddrP (P1(_,_,_,j,m)) = SOME (j,m)
  | outerAddrP (P2(_,_,_,j,m)) = SOME (j,m)
  | outerAddrP (O3(_,_,_,j,m)) = SOME (j,m)
  | outerAddrP (P4(_,_,j,m))   = SOME (j,m)
  | outerAddrP (P5(_,j,m))   = SOME (j,m)
  | outerAddrP (O6(_,j,m))   = SOME (j,m)
  | outerAddrP _ = NONE ;

fun topIndexO (P7(_,_,_,t))    = SOME t
  | topIndexO (O8(_,t))      = SOME t
  | topIndexO (P9(_,_,_,t))  = SOME t
  | topIndexO (O10(_,_,_,t)) = SOME t
  | topIndexO _ = NONE ;

fun topIndexP (O7(_,_,_,t))    = SOME t
  | topIndexP (P8 t)         = SOME t
  | topIndexP (O9(_,_,_,t))  = SOME t
  | topIndexP (P10(_,_,_,t)) = SOME t
  | topIndexP _ = NONE ;

fun rhsAddrP (O9(_,j,m,t))  = SOME (j,m)
  | rhsAddrP (P10(_,j,m,t)) = SOME (j,m)
  | rhsAddrP _ = NONE ;

exception UnexpectedWriteIndex of index
exception UnexpectedReadIndex of index
exception UnexpectedTopIndex
exception UnexpectedHistory
exception TrackingError  (* problem in history, bug somewhere *)
exception TrackingBug    (* bug in tracking code *)
exception ReadAssocError
exception NonBodyMove ;
exception UnknownObjectId ;
exception MoveBug ;

type History = (threadPmove * threadOmove * objectId) list ; 

type Admin = {
     history : History,
         (* list of (pm,om) pairs, most recent first *)
     nextTopIndex : topIndex,
         (* index counter for the topmost bang *)
     lastWriteMoves : (objectId * label, (index*label) option) alist,
         (* last write move in O4, if any *)
     groundStates : (objectId * label, SG) alist,
         (* ground type component of current state *)
     birthPoints : (objectId, topIndex) alist,
         (* topmost index of P8 move corresponding to object creation *)
     readAssoc : (index*label*index*label, (index*label) option) alist
         (* association of "read" return points (col 1) with 
            "write" call points (col 4), or with initial state.
            Inner index of write call points omitted as it's always 0. 
            Inner label of write call points is same as for read points *)
}

(* Streams for justification sequences w.r.t. top index *)

datatype ('o,'p) Pstream = Empty 
			 | Pstream of 'p * (unit -> ('o,'p) Ostream)
     and ('o,'p) Ostream = Ostream of 'o * (unit -> ('o,'p) Pstream) ;

fun Ohead (Ostream (om,_)) = om ;
fun Ptail (Ostream (_,rest)) = rest() ;


(* Easy admin operations: *)

fun addMovePair (stuff as (pm,_,_)) (admin:Admin) : Admin = {
    history = stuff :: #history admin,
    nextTopIndex = (case topIndexP pm of 
			NONE => #nextTopIndex admin 
		      | SOME _ => inc (#nextTopIndex admin)),
    lastWriteMoves = #lastWriteMoves admin,
    groundStates = #groundStates admin,
    birthPoints = #birthPoints admin,
    readAssoc = #readAssoc admin
    } ;

fun recordWrite objf sg jm (admin:Admin) = {
    lastWriteMoves = updateAlist (objf, SOME jm) (#lastWriteMoves admin),
    groundStates = updateAlist (objf, sg) (#groundStates admin),
    history = #history admin,
    nextTopIndex = #nextTopIndex admin,
    birthPoints = #birthPoints admin,
    readAssoc = #readAssoc admin
    } ;

exception LastWriteError ;
fun getLastWriteMove objf (admin:Admin) = 
    errorLookup (fn _ => LastWriteError) (#lastWriteMoves admin) objf ;

exception GroundError ;
fun getGroundState objf (admin:Admin) = 
    errorLookup (fn _ => GroundError) (#groundStates admin) objf ;

exception BirthError ;
fun getBirthPoint obj (admin:Admin) = 
    errorLookup (fn _ => BirthError) (#birthPoints admin) obj ;

fun getReadAssoc ifjm (admin:Admin) = 
    errorLookup (fn _ => ReadAssocError) (#readAssoc admin) ifjm ;

fun addReadAssoc obj f (i,j,m) (admin:Admin) = {
    (* associates current last write move for f (if any) with (i,f,j,m) *)
    readAssoc = ((i,f,j,m), getLastWriteMove (obj,f) admin) :: #readAssoc admin,
    history = #history admin,
    nextTopIndex = #nextTopIndex admin,
    lastWriteMoves = #lastWriteMoves admin,
    groundStates = #groundStates admin,
    birthPoints = #birthPoints admin
    } ;

fun newObjectId () = ref () ;

fun installNewObject (obj,t,sg_all) (admin:Admin) = {
    birthPoints = (obj,t) :: #birthPoints admin,
    groundStates = map' sg_all (fn (f,sg) => ((obj,f),sg)) 
		   @ #groundStates admin,
    lastWriteMoves = map' sg_all (fn (f,_) => ((obj,f),NONE)) 
		     @ #lastWriteMoves admin,
    history = #history admin,
    nextTopIndex = #nextTopIndex admin,
    readAssoc = #readAssoc admin
    } ;

fun OtopIndexHistory [] = raise ReadAssocError
  | OtopIndexHistory ((_,om,_)::rest) =
    case topIndexO om of 
	SOME t => (t,rest) | NONE => OtopIndexHistory rest ;

fun findTopIndex t [] = NONE
  | findTopIndex t ((pm,om,_)::rest) =
    if topIndexP pm = SOME t then SOME (pm,om,rest)
    else findTopIndex t rest ;

fun OmoveHistoryAt _ [] = raise ReadAssocError
  | OmoveHistoryAt (j,m) ((_,om,_)::rest) =
    if outerAddrO om = SOME (j,m) then (om,rest)
    else OmoveHistoryAt (j,m) rest ;

fun fetchObj om [] = raise UnexpectedHistory
  | fetchObj om (hist as ((_,_,obj)::rest)) =
    (case topIndexO om of
	 NONE => obj
       | SOME t => 
	 (case List.find (fn (pm,_,_) => topIndexP pm = SOME t) hist of
               NONE => raise UnexpectedHistory
	     | SOME (_,_,obj') => obj')) ;


(* Less trivial admin operations, using streams *)

fun PstreamOf [] = Empty
  | PstreamOf ((pm,om,_)::rest) =
    Pstream (pm, fn () => Ostream (om, fn () => PstreamOf rest)) ;

fun topPstreamOf t [] = Empty
  | topPstreamOf t ((pm,om,obj) :: rest) =
    if topIndexP pm = SOME t then
        (case pm of P8 _ => Empty 
                  | _ => Pstream (pm, fn () => topOstreamOf om rest))
    else topPstreamOf t rest
and topOstreamOf om rest =
    case topIndexO om of
	SOME t => Ostream (om, fn () => topPstreamOf t rest)
      | NONE => (case rest of
	     [] => raise UnexpectedHistory
           | ((_,om',_) :: rest') => topOstreamOf om' rest') ;

fun hStream (admin:Admin) = PstreamOf (#history admin) ;

(* Similar thing but gives the Omove preceding the Pmove
   rather than the one on the justification path: *)
fun topPstreamOf' t [] = Empty
  | topPstreamOf' t ((pm,om,obj) :: rest) =
    if topIndexP pm = SOME t then
        (case pm of P8 _ => Empty 
                  | _ => Pstream (pm, fn () => 
                            Ostream (om, fn () => topPstreamOf'' om rest)))
    else topPstreamOf' t rest
and topPstreamOf'' om rest =
    case topIndexO om of
	SOME t => topPstreamOf' t rest
      | NONE => (case rest of
	     [] => raise UnexpectedHistory
           | ((_,om',_) :: rest') => topPstreamOf'' om' rest') ;

fun topStream t (admin:Admin) = topPstreamOf' t (#history admin) ;

(* A couple of rhs operations, working on topStreams: *)
fun nextRhsIndex m Empty = 1
  | nextRhsIndex m (Pstream (pm,rest)) =
    case rhsAddrP pm of
	NONE => nextRhsIndex m (Ptail (rest()))
      | SOME (j,m') =>
	if m'=m then j+1 else nextRhsIndex m (Ptail (rest())) ;

(* Move pair whose P-move is the O9 or P10 move labelled (j,m): *)
fun rhsMovePairAt jm Empty = raise UnexpectedHistory
  | rhsMovePairAt jm (Pstream (pm, rest)) =
      if rhsAddrP pm = SOME jm then (pm, Ohead(rest()))
      else rhsMovePairAt jm (Ptail (rest())) ;

(* Operations working on hStreams: *)
fun bodyMoveHistoryAt jm Empty = raise TrackingError
  | bodyMoveHistoryAt jm (Pstream (pm,rest)) = findOuter jm (rest())
and findOuter jm (Ostream (om,rest)) =
    if outerAddrO om = SOME jm then (om,rest())
    else bodyMoveHistoryAt jm (rest()) ;

fun backToMethod m Empty = raise UnexpectedHistory
  | backToMethod m (hist as Pstream (pm,rest)) =
       if (case outerAddrP pm of SOME (_,m') => m'=m | NONE => false)
       then hist else backToMethod m (Ptail (rest())) ;

fun chaseToMovePair m Property h =
    chaseToMovePair' m Property (backToMethod m h) 
and chaseToMovePair' m Property Empty = NONE
  | chaseToMovePair' m Property (Pstream (pm, rest)) =
    let val Ostream (om,rest') = rest() in
	if Property pm then SOME (pm,om,rest'())
	else (case outerAddrP pm of
		  SOME (0,_) => NONE
		| SOME jm => let val (_,rest'') = findOuter jm (rest()) 
			     in chaseToMovePair m Property rest''
			     end
		| NONE => raise NonBodyMove) 
    end ;
(* Maybe clearer if we pull out the justification substream for m? 
   Not sure: below we sometimes use the move om, which doesn't lie on the
   justification path... *)

fun chaseToPmove m P h = case chaseToMovePair m P h of
        SOME (pm,_,_) => SOME pm | NONE => NONE ;

fun nextWriteIndex m history = 
    (* next inner index for "write" bang along the justification path 
       chased back from last P-move *)
    (case chaseToPmove m 
	      (fn O3 _ => true | P4 _ => true | _ => false) 
              history of
        SOME (O3(_,i,_,_,_)) => i+1 
      | SOME (P4(i,_,_,_)) => i+1
      | NONE => 1 | _ => raise TrackingBug) ;

fun nextReadIndex m history = 
    (* similarly *)
    (case chaseToPmove m 
              (fn P1 _ => true | P2 _ => true | _ => false) 
              history of
        SOME (P1(_,i,_,_,_)) => i+1 
      | SOME (P2(_,i,_,_,_)) => i+1
      | NONE => 1 | _ => raise TrackingBug) ;


(* The main operation on moves *)

fun 
    (* Creating a new object: *)
    threadMove (O8 (sg_all,t)) admin = 
    if t = wrap 0 then
	let val (t',obj) = (#nextTopIndex admin, newObjectId())
	in ((P8 t', installNewObject (obj,t',sg_all) admin), obj)
	end
    else raise UnexpectedTopIndex

    (* All other enquiries: fetch relevant objectId *)
  | threadMove om admin = 
    let val obj = fetchObj om (#history admin)
    in (threadMove' om obj admin, obj)
    end

and
  (* Copying moves in B_m: *)
    threadMove' (O10 (bo,0,m,t)) obj admin = (O6 (bo,0,m), admin)
  | threadMove' (O10 (bo,j,m,t)) obj admin =
      (case rhsMovePairAt (j,m) (topStream t admin) of
           (P10 _, P6 (_,j',m')) => (O6 (bo,j',check m' m), admin)
         | _ => raise UnexpectedHistory)

  | threadMove' (P6 (bp,j,m)) obj admin =
      let val (t,rest) = OtopIndexHistory (#history admin)
          val j' = nextRhsIndex m (topPstreamOf t rest)
      in (P10 (bp,j',m, #nextTopIndex admin), admin)
      end

  (* Copying moves in A_m, similarly: *)
  | threadMove' (O5 (ao,j,m)) obj admin =
      let val (t,rest) = OtopIndexHistory (#history admin)
          val j' = nextRhsIndex m (topPstreamOf t rest) in
          (O9 (ao, j', m, #nextTopIndex admin), admin)
      end
      (* NB: a safety violation should be detected as a parity violation
         within A_m -o B_m (columns 9 and 10). Check this! 
         (Unless we do an explicit check, may as well merge A_m, B_m ?) *)

  | threadMove' (P9 (ap,j,m,t)) obj admin =  
      (* j can't be 0 here *)
      (case rhsMovePairAt (j,m) (topStream t admin) of
           (O9 _, O5 (_,j',m')) => (P5 (ap,j',check m' m), admin)
	 | _ => raise UnexpectedHistory)

  (* Calling "write": *)
  | threadMove' (O4 (g,i,f,j,m)) obj admin =
    if i=0 then 
        (P4 (nextWriteIndex m (hStream admin), f, j, m), 
	 recordWrite (obj,f) g (j,m) admin)
    else raise UnexpectedWriteIndex i

  (* Calling "read": set up connection with initial state or with 
     most recent argument to "write" *)
  | threadMove' (O1 (i,f,j,m)) obj admin =
    if i=0 then 
        let val ri = nextReadIndex m (hStream admin)
            val sg = getGroundState (obj,f) admin
        in  (P1 (sg,ri,f,j,m), 
             addReadAssoc obj f (ri,j,m) admin)
        end
    else raise UnexpectedReadIndex i 

  (* Copycat between read result and initial or written state:
     based on chasing J-pointers *)
  | threadMove' (O2 (so,i,f,j,m)) obj admin =
    ((case chaseToMovePair m
	       (fn P1(_,i',f',_,_) => i'=i andalso f'=f
		 | P2(_,i',f',_,_) => i'=i andalso f'=f
		 | _ => false )
	       (hStream admin) of
         SOME (P1 (g,i',f',j',m'), _, _) => 
         if (i',f',m') = (i,f,m) then
  	     (case getReadAssoc (i',f',j',m') admin of
	            NONE => O7 (so,0,f,#nextTopIndex admin)
		  | SOME (j'',m'') => 
                       let val i'' = 
	                   (case OmoveHistoryAt (j'',m'')
				 (#history admin) of 
			    (O4 _, rest) => 
               		         nextWriteIndex m'' (PstreamOf rest) 
			  | _ => raise ReadAssocError)
	               in O3 (so,i'',f,j'',m'')
                       end)
         else raise TrackingBug 
       | SOME (P2 _, P7 (sp,h,f',_), _) => O7 (so,h,f',#nextTopIndex admin)
       | SOME (P2 _, P3 (sp,i',f',j',m'), rest) => 
	      O3 (so, nextWriteIndex m' rest, f', j', m')
       | _ => raise UnexpectedHistory) ,
     admin)

  | threadMove' (P7 (sp,h,f,t)) obj admin =  
    (* always responds to most recent move in justification path *) 
    (case findTopIndex t (#history admin) of
	 SOME (_, O2 (so,i,f',j,m), rest) =>
	      (P2 (sp, nextReadIndex m (PstreamOf rest), f', j, m), admin)
       | _ => raise UnexpectedHistory)

  | threadMove' (P3 (sp,i,f,j,m)) obj admin =
    (case chaseToMovePair m 
    (* chase needed here if we store e.g. suspended nested calls *)
             (fn O3 (_,ir,_,_,_) => ir=i | _ => false) 
             (hStream admin) of  
         SOME (_, O2 (so,i',f',j',m'), rest) => 
              (P2 (sp, nextReadIndex m' rest, f', j', m'), admin)
       | _ => raise UnexpectedHistory)

  | threadMove' (O8 _) _ _ = raise MoveBug
  ;


(* Thread strategy, dependent on an initial ground state *)

datatype threadStrat = fold of threadOmove -> threadPmove * threadStrat ;

val adminStore = ref (NONE : Admin option) ;  (* for debugging *)

fun threadStrat' (admin:Admin) : threadStrat =
    fold (fn om:threadOmove =>
	     case threadMove om admin of ((pm,admin'),obj) => 
                 let val admin'' = addMovePair (pm,om,obj) admin'
                 in  adminStore := SOME admin'' ;
 		     (pm, threadStrat' admin'')
                 end)

val threadStrat = threadStrat' {
    history = [], nextTopIndex = wrap 1, lastWriteMoves = [], 
    groundStates = [], birthPoints = [], readAssoc = []} ;


(* For interactive testing *)

local fun diverge() = diverge() in
val dummyStrat = fold (fn om => diverge())
end ;

val theStrat = ref dummyStrat
val dummyObj = ref () 
val theOuterIndex = ref (0, dummyObj)
val theObjects = ref ([] : unit ref list) ;
fun addObject obj [] = [obj]
  | addObject obj (obj'::rest) = 
    if obj'=obj then obj'::rest else obj'::addObject obj rest
fun object i = List.nth (!theObjects, i) ;
fun install s = 
    (theStrat := s ; theOuterIndex := (0, dummyObj) ; theObjects := []) ;
fun play Op data =
    let val fold f = !theStrat ;
        val (pm, s') = f (Op data)
    in 
       theStrat := s' ; 
       pm
    end ;

end (* functor Thread *) ;


(* Tying the above into our main game model 
   via boring encodings/decodings of moves. 
   Also a bit of massaging to give THREAD a language-denotable type. *)

signature CBN_THREAD = sig
   type morphism
   val THREAD : string list -> morphism
end ;

functor CBN_Thread (
   structure CBN_Games : CBN_GAMES ;
   val prodDecode : CBN_Games.ground -> (string * CBN_Games.ground) list
   val prodEncode : (string * CBN_Games.ground) list -> CBN_Games.ground
   ) : CBN_THREAD = struct

structure Thread0 = Thread (
   type SG = CBN_Games.ground ;
   type SO = CBN_Games.omove ;
   type SP = CBN_Games.pmove ;
   type AO = CBN_Games.omove ;
   type AP = CBN_Games.pmove ;
   type BO = CBN_Games.omove ;
   type BP = CBN_Games.pmove
   ) ;

local 
    open Pervasives
    open CBN_Games 
    open Thread0
in

type morphism = morphism ;

fun blanks mths = prodEncode (map (fn m => (m,blank)) mths)

fun codePmove mths (P1 (sg,i,f,j,m)) =
      Source (ProdO (m, TagO (j, LeftO (ProdP (f, ProdP ("read", 
	  TagP (i, RightP (BasicP sg))))))))
  | codePmove _ (P2 (sp,i,f,j,m)) =
      Source (ProdO (m, TagO (j, LeftO (ProdP (f, (ProdP ("read", 
          TagP (i, RightP (MainP sp)))))))))
  | codePmove _ (O3 (so,i,f,j,m)) =
      Source (ProdO (m, TagO (j, LeftO (ProdP (f, (ProdP ("write", 
          TagP (i, LeftP so))))))))
  | codePmove _ (P4 (i,f,j,m)) =
      Source (ProdO (m, TagO (j, LeftO (ProdP (f, (ProdP ("write", 
          TagP (i, RightP (BasicP (blanks []))))))))))
  | codePmove _ (P5 (ap,j,m)) =
      Source (ProdO (m, TagO (j, RightO (LeftO ap))))
  | codePmove _ (O6 (bo,j,m)) =
      Source (ProdO (m, TagO (j, RightO (RightO bo))))
  | codePmove _ (O7 (so,h,f,t)) =
      Target (TagP (unwrap t, LeftP (ProdO (f, (TagO (h, so))))))
  | codePmove mths (P8 t) =
      Target (TagP (unwrap t, RightP (BasicP (blanks mths))))
  | codePmove _ (O9 (ao,j,m,t)) =
      Target (TagP (unwrap t, RightP (MainP (ProdP (m, TagP (j, 
          LeftP ao))))))
  | codePmove _ (P10 (bp,j,m,t)) =
      Target (TagP (unwrap t, RightP (MainP (ProdP (m, TagP (j, 
          RightP bp))))))

exception IllegalThreadOmove of (pmove,omove) Mmove

fun decodeOmove 
      (Source (ProdP (m, TagP (j, LeftP (ProdO (f, (ProdO ("read", 
	  TagO (i, RightO (BasicO blank))))))))))
          (* minor wrinkle: "blank" is a pattern variable here *)
       = O1 (i,f,j,m)
  | decodeOmove 
      (Source (ProdP (m, TagP (j, LeftP (ProdO (f, (ProdO ("read", 
          TagO (i, RightO (MainO so))))))))))
       = O2 (so,i,f,j,m)
  | decodeOmove 
      (Source (ProdP (m, TagP (j, LeftP (ProdO (f, (ProdO ("write", 
          TagO (i, LeftO sp)))))))))
       = P3 (sp,i,f,j,m)
  | decodeOmove 
      (Source (ProdP (m, TagP (j, LeftP (ProdO (f, (ProdO ("write", 
          TagO (i, RightO (BasicO sg))))))))))
       = O4 (sg,i,f,j,m)
  | decodeOmove (Source (ProdP (m, TagP (j, RightP (LeftP ao)))))
       = O5 (ao,j,m)
  | decodeOmove (Source (ProdP (m, TagP (j, RightP (RightP bp)))))
       = P6 (bp,j,m)
  | decodeOmove (Target (TagO (t, LeftO (ProdP (f, TagP (h, sp))))))
       = P7 (sp,h,f,wrap t)
  | decodeOmove (Target (TagO (t, RightO (BasicO sg))))
       = O8 (prodDecode sg, wrap t)
  | decodeOmove 
      (Target (TagO (t, RightO (MainO (ProdO (m, TagO (j, 
          LeftO ap)))))))
       = P9 (ap,j,m,wrap t)
  | decodeOmove 
      (Target (TagO (t, RightO (MainO (ProdO (m, TagO (j, 
          RightO bo)))))))
       = O10 (bo,j,m,wrap t)
  | decodeOmove om = (print ("Culprit: " ^ MomoveToS om ^ "\n") ;
		      raise IllegalThreadOmove om)

(* 
(* Not strictly needed, but possibly handy: *)
(* NOW OUT OF DATE ANYWAY: f's need adding *)

fun codeOmove (O1 (i,j,m)) =
      Source (ProdP (m, TagP (j, LeftP (ProdO ("read", 
	  TagO (i, BasicO blank))))))
  | codeOmove (O2 (so,i,j,m)) =
      Source (ProdP (m, TagP (j, LeftP (ProdO ("read", 
          TagO (i, MainO so))))))
  | codeOmove (P3 (sp,i,j,m)) =
      Source (ProdP (m, TagP (j, LeftP (ProdO ("write", 
          TagO (i, LeftO sp))))))
  | codeOmove (O4 (sg,i,j,m)) =
      Source (ProdP (m, TagP (j, LeftP (ProdO ("write", 
          TagO (i, RightO (BasicO sg)))))))
  | codeOmove (O5 (ao,j,m)) =
      Source (ProdP (m, TagP (j, RightP (LeftP ao))))
  | codeOmove (P6 (bp,j,m)) =
      Source (ProdP (m, TagP (j, RightP (RightP bp))))
  | codeOmove (P7 (sp,h,t)) =
      Target (TagO (unwrap t, LeftO (TagP (h, sp))))
  | codeOmove (O8 (sg,t)) =
      Target (TagO (unwrap t, RightO (BasicO sg)))
  | codeOmove (P9 (ap,j,m,t)) =
      Target (TagO (unwrap t, RightO (MainO (ProdO (m, TagO (j, 
          LeftO ap))))))
  | codeOmove (O10 (bo,j,m,t)) =
      Target (TagO (unwrap t, RightO (MainO (ProdO (m, TagO (j, 
          RightO bo))))))

(* ...and maybe also decodePmove. *)
*)

fun translate mths (fold f) : morphism =
    strategy (fn om => 
	case f (decodeOmove om) of (pm',rest) =>
		(codePmove mths pm', translate mths rest)) 

(* Massaging to a language-denotable type with an extra "\bot" *)

infix THEN

val POST_FLATTEN : morphism =
    CURRY ("#1","#2") (UNCURRY ("#1","#2") ID THEN FLATTEN)

fun THREAD methods : morphism = 
    (* can perhaps avoid having to supply the list of method names *)
    PRODUCT (map' methods (fn m => (m, BANG POST_FLATTEN)))
    THEN translate methods threadStrat 

end 

end (* functor CBN_Thread *) ;
