
structure Types = struct

type var = string
type label = string
type linopt = bool

type errpos = int * int
exception Unimplemented of int
exception Untypeable of errpos * string
exception Unsafe of errpos * string
exception Impossible of string

(* ASTs for Lingay underlying types, and associated operations *)
		  		  
datatype ty =  BoolTy | IntTy | ForestTy
	     | ProdTy of (label * ty) list * var list
	     | SumTy of (label * ty) list
	     | BangTy of ty
	     | FunTy of ty * ty
	     | RecTy of var * ty
	     | ClassImplTy of linopt * ty * ty * ty
	     | NamedTy of var
             | MergeTy of ty * ty

fun fetchRest v [] = NONE
  | fetchRest v ((v',t)::rest) =
    if v'=v then SOME (t,rest) else fetchRest v rest

fun componentsOf _ (ProdTy (complist,_)) = complist
  | componentsOf _ (SumTy complist) = complist
  | componentsOf tenv (NamedTy v) = 
    (case fetchRest v tenv of SOME (ty,tenv') => componentsOf tenv' ty
			    | NONE => [])
  | componentsOf tenv (MergeTy (t1,t2)) =
    (componentsOf tenv t1) @ (componentsOf tenv t2)
  | componentsOf _ _ = []

fun fst (x,y) = x and snd (x,y) = y

fun classDataOf tenv (ClassImplTy (lo,ty1,ty2,_)) =
    (map fst (componentsOf tenv ty1), map fst (componentsOf tenv ty2), lo)
  | classDataOf tenv (NamedTy v) =
    (case fetchRest v tenv of SOME (ty,tenv') => classDataOf tenv' ty
			    | NONE => ([],[],false))
  | classDataOf _ _ = ([],[],false)

local open Pervasives; exception TrivGroundError in
fun trivGroundFor tenv (FunTy _) = blankv
  | trivGroundFor tenv (BangTy t) = trivGroundFor tenv t
  | trivGroundFor tenv (ProdTy (tys,_)) =
        prodv (map' tys (fn (l,t) => (l, trivGroundFor tenv t)))
  | trivGroundFor tenv (SumTy [(l,t)]) = injv (l, trivGroundFor tenv t)
  | trivGroundFor _ (ForestTy) = foldv blankv
  | trivGroundFor tenv (RecTy (v,t)) = foldv (trivGroundFor tenv t)
  | trivGroundFor _ (ClassImplTy _) = prodv [("#1",blankv),("#2",blankv)]
  | trivGroundFor tenv (NamedTy v) = 
    (case fetchRest v tenv of SOME (ty,tenv') => trivGroundFor tenv' ty
			    | NONE => raise TrivGroundError)
  | trivGroundFor _ _ = raise TrivGroundError
end

fun concatWith sep [] = ""
  | concatWith sep [h] = h
  | concatWith sep (h::t) = h ^ sep ^ (concatWith sep t)

fun tytostring t =
  let fun cts l = concatWith ", " (map (fn (k,t) => k ^ ": " ^ (tytostring t)) l)
      fun ctt l = concatWith ", " (map (fn (k,t) => k ^ " of " ^ (tytostring t)) l)
  in
  case t of BoolTy => "bool"
      | IntTy => "int"
      | ForestTy => "forest"
      | ProdTy(ts,more) => "{" ^ (cts ts) ^ "}" ^ (concat (map (fn x => " ** "^x) more))
      | SumTy(ts) => "[|" ^ (ctt ts) ^ "|]"
      | BangTy(FunTy(t1,t2)) => "(" ^ tytostring t1 ^ " -> " ^ tytostring t2 ^ ")"
      | BangTy t => "!" ^ tytostring t
      | FunTy(t1,t2) => "(" ^ tytostring t1 ^ " -o " ^ tytostring t2 ^ ")"
      | RecTy(x,t) => "(rectype " ^ x ^ " => " ^ tytostring t ^ ")"
      | ClassImplTy(lin,t1,t2,t3) => (if lin then "linear " else "") ^ "classimpl " 
                ^ (tytostring t1) ^ ", " ^(tytostring t2)^ ", "^(tytostring t3)^" end"
      | NamedTy x => x
      | MergeTy (t1,t2) => "(" ^ tytostring t1 ^ " ** " ^ tytostring t2 ^ ")"
  end

(* Type status *)

datatype ty_status = 
	 Ordinary | Reusable | Ground

(* elt of list l1 duplicated in l2 *)
fun duplicates (f: 'a -> ''b) (l1: 'a list) (l2: 'a list) = 
    List.filter (fn k => List.exists (fn k' => (f k)=(f k')) l2) l1 

(* label duplicated in single list *)
fun duplicates1 (f: 'a -> ''b) ((h::t): 'a list) = 
    if (List.exists (fn x => f x=f h) t) 
    then (f h)::(duplicates1 f t) else duplicates1 f t
  | duplicates1 f [] = []

fun worse_status (NONE, _) = NONE
  | worse_status (_, NONE) = NONE
  | worse_status (SOME Ordinary, SOME _) = SOME Ordinary
  | worse_status (SOME _, SOME Ordinary) = SOME Ordinary
  | worse_status (SOME Reusable, SOME _) = SOME Reusable
  | worse_status (SOME _, SOME Reusable) = SOME Reusable
  | worse_status (SOME Ground, SOME Ground) = SOME Ground

exception MergeTypeError

fun disjoint l1 l2 = List.all (fn x => List.all (fn y => y<>x) l2) l1 ;

fun resolvety (pos:errpos) env (NamedTy x) = 
    (case fetchRest x env of 
	 SOME (ty,env') => resolvety pos env' ty (* recursive problem *)
       | NONE => NamedTy x)
  | resolvety pos env BoolTy    = BoolTy
  | resolvety pos env IntTy     = IntTy
  | resolvety pos env ForestTy  = ForestTy
  | resolvety pos env (ProdTy(tys,ext)) = ProdTy(map (fn (l,t) => (l, resolvety pos env t)) tys,ext)
  | resolvety pos env (SumTy tys) = SumTy(map (fn (l,t) => (l, resolvety pos env t)) tys)
  | resolvety pos env (BangTy t) = BangTy(resolvety pos env t)
  | resolvety pos env (FunTy(t1,t2)) = FunTy(resolvety pos env t1, resolvety pos env t2)
  | resolvety pos env (RecTy(x,t)) = 
    if List.all (fn (y,_) => y<>x) env then RecTy(x, resolvety pos env t)
    else raise Untypeable (pos, "Bound type variable already in scope: " ^x)
  | resolvety pos env (ClassImplTy(lin,t1,t2,t3)) = ClassImplTy(lin,resolvety pos env t1, resolvety pos env t2, resolvety pos env t3)
  | resolvety pos env (MergeTy(t1,t2)) =
    (case (resolvety pos env t1, resolvety pos env t2) of
	 (ProdTy(tys1,ext1), ProdTy(tys2,ext2)) =>
	      if disjoint (map fst tys1) (map fst tys2) then ProdTy (tys1@tys2, ext1@ext2)
              else raise Untypeable (pos, "Merged record type has repeated labels: " ^ 
					  (concatWith ", " (map fst tys1)) ^ " vs. " ^ (concatWith ", " (map fst tys2)))
       | _ => raise Untypeable (pos, "Attempt to merge non-product types"))

(* Obsolete, not updated with RecTy and MergeTy: *)
fun rowvar_notin x (BoolTy) = true
  | rowvar_notin x (IntTy) = true
  | rowvar_notin x (ForestTy) = true
  | rowvar_notin x (NamedTy name) = raise Impossible ("Named type found where type should be fully resolved (rowvar_notin): "^name)
  | rowvar_notin x (ProdTy(tys,vars)) = List.all (fn y=>y<>x) vars 
					andalso List.all (rowvar_notin x) (map snd tys)
  | rowvar_notin x (SumTy(tys)) = List.all (rowvar_notin x) (map snd tys)
  | rowvar_notin x (BangTy(t)) = rowvar_notin x t
  | rowvar_notin x (FunTy(t1,t2)) = rowvar_notin x t1 andalso rowvar_notin x t2
  | rowvar_notin x (ClassImplTy(lin,t1,t2,t3)) = List.all (rowvar_notin x) [t1,t2,t3]
  | rowvar_notin _ _ = false (* dummy *)

fun check_unique_labels pos labtys =
    let val dups = duplicates1 fst labtys
    in 
        if dups = [] then () else raise Untypeable(pos, "Product/sum type has non-unique labels: "^(concatWith ", " dups))
    end

(* Compute the best status descriptor of some type iff it is well formed *)
fun status senv pos (BoolTy) = SOME Ground
  | status senv pos (IntTy) = SOME Ground
  | status senv pos (ForestTy) = SOME Ordinary
  | status senv pos (NamedTy x) = 
    (case Pervasives.optLookup senv x of SOME st => SOME st 
			   | NONE => raise Impossible ("Named type found where type should be fully resolved (status): "^x))
  | status senv pos (MergeTy _) = raise Impossible ("Merged type found where type should be fully resolved (status) ")
  | status senv pos (ProdTy(tys,ext)) = 
    let val () = check_unique_labels pos tys
        val stat = foldl worse_status (SOME Ground) (map ((status senv pos) o snd) tys)
    in case ext of [] => stat
		 | _ => if is_ordinary stat andalso List.all (fn z => rowvar_notin z (ProdTy(tys,[]))) ext
			then SOME Ordinary (* stat may be better than ordinary, downgrade *)
			else NONE
    end
  | status senv pos (SumTy []) = NONE  (* empty type excluded for now *)
  | status senv pos (SumTy tys) = 
    (check_unique_labels pos tys;
     foldl worse_status (SOME Ground) (map ((status senv pos) o snd) tys))
  | status senv pos (BangTy t) = (case status senv pos t of SOME Ground => SOME Ground
                    | SOME _ => SOME Reusable
                    | NONE => NONE)
  | status senv pos (FunTy(t1,t2)) = if is_ordinary (status senv pos t1)
                   andalso is_ordinary (status senv pos t2)
                then SOME Ordinary else NONE
  | status senv pos (RecTy(x,t)) = 
      (case status ((x,Ground)::senv) pos t of
          SOME st1 => status ((x,st1)::senv) pos t
        | NONE => NONE)
  | status senv pos (ClassImplTy(lin,t1,t2,t3)) = 
      case lin of true =>  if is_ordinary (status senv pos t1) 
			      andalso is_reusable (status senv pos t2) 
			      andalso is_ordinary (status senv pos t3)
				      (* is_ground to agree with formal definition *)
			      andalso ground_class_form pos (t1,t2) 
			   then SOME Reusable else NONE
        | false => if is_reusable (status senv pos t1) andalso is_reusable (status senv pos t2) andalso is_ordinary (status senv pos t3) 
                  andalso class_form (t1, t2) then SOME Reusable else NONE
and statusE pos t = case status [] pos t of SOME s => s | NONE => raise Untypeable(pos, "Invalid type: "^(tytostring t))
(* FIX: row vars *)
and class_form  (ProdTy(_,[]), ProdTy(tys,[])) = List.all (fn (k, BangTy(FunTy(t1,t2))) => true | _ => false) tys
  | class_form  _ = false
and ground_class_form pos (ProdTy(_,[]), ProdTy(tys,[])) = 
    List.all (fn (k, BangTy(FunTy(t1,t2))) => is_ground (status [] pos t1) 
               | _ => false) tys
  | ground_class_form pos _  = false

and is_ordinary (SOME _) = true 
  | is_ordinary NONE = false
and is_reusable (SOME Reusable) = true
  | is_reusable (SOME Ground) = true
  | is_reusable _ = false
and is_ground (SOME Ground) = true
  | is_ground _ = false

fun is_reusable_ty pos tenv ty = 
    is_reusable (status [] pos (resolvety pos tenv ty))

fun groundless (IntTy) = false
  | groundless (BoolTy) = false
  | groundless (ForestTy) = true
  | groundless (ProdTy(tys,[])) = List.all (groundless o snd) tys
  | groundless (ProdTy(tys,h::t)) = false
  | groundless (SumTy tys) = length tys = 1 andalso List.all (groundless o snd) tys 
  | groundless (BangTy t) = groundless t
  | groundless (FunTy(t1,t2)) = true
  | groundless (RecTy(x,t)) = groundless t
  | groundless (ClassImplTy(lin,t1,t2,t3)) = true
  | groundless (NamedTy _) = true  (* trick! *)
  | groundless (MergeTy _) = raise Impossible "Unexpected merged type found in groundless"

infix <:
infix </:
infix =:=
infix <:>
(* inefficient *)
fun subtype sbenv (t1,t2) =
    if t1 = t2 then true
    else case (t1,t2) of 
             (NamedTy x1, NamedTy x2) => List.exists (fn p => p=(x1,x2)) sbenv
           | (ProdTy (tys1,x), ProdTy (tys2,y)) => 
             x = y andalso List.all (fn (k,t) => groundless t orelse List.exists (fn (k',t') => k'=k andalso subtype sbenv (t,t')) tys2) tys1
                   andalso List.all (fn (k', t') => List.exists (fn (k,t) => k=k' andalso subtype sbenv (t,t')) tys1) tys2
           | (SumTy tys1, SumTy tys2) => List.all (fn (k, t) => List.exists (fn (k',t') => k=k' andalso subtype sbenv (t,t')) tys2) tys1
           | (BangTy t1', BangTy t2') => subtype sbenv (t1', t2')
           | (FunTy (t11,t12), FunTy (t21,t22)) => subtype sbenv (t21,t11) andalso subtype sbenv (t12,t22) 
           | (RecTy (x1,t1), RecTy (x2,t2)) => subtype ((x1,x2)::sbenv) (t1,t2)
           | (ClassImplTy(lin1,ty11,ty12,ty13), ClassImplTy(lin2,ty21,ty22,ty23)) => 
             lin1 = lin2 andalso ty11 = ty21 andalso ty12 = ty22  andalso subtype sbenv (ty23,ty13)
           | _ => false
fun t1 <: t2 = subtype [] (t1,t2)
fun t1 </: t2 = not (t1 <: t2)
fun t1 =:= t2 = t1 <: t2 andalso t2 <: t1
fun t1 <:> t2 = not (t1 =:= t2)


(* p1 in p2? i.e. does there exist p such that p1+p = p2 ?  (equiv oplus)*)
fun prod_inclusion pos (ProdTy(tys1,vars1)) (ProdTy(tys2,vars2)) = 
    List.take(vars2, List.length vars1) = vars1 
    andalso List.all (fn (k', t') => List.exists (fn (k,t) => k=k' andalso  t =:= t') tys1) tys2
  | prod_inclusion pos _ _ = false

(* p1 + p2 : combine products p1 and p2 with distinct labels and row variables *)
fun prod_plus pos (tys1,vars1) (tys2,vars2) = 
    let
        val dups = duplicates fst tys1 tys2
        val () = if dups = [] then () else
                     raise Untypeable(pos,"Labels should be distinct: " 
                                          ^ (concatWith ", " (map (fn (k,_)=>k) dups)))
        val dups = duplicates (fn x=>x) vars1 vars2
        val () = if dups = [] then () else
                     raise Untypeable(pos,"Row variables should be distinct: " 
                                          ^ (concatWith ", " dups))
    in (tys1@tys2, vars1@vars2) end

(* p1 \oplus p2 : combine compatible products p1 and p2, favouring p2. *)
fun prod_oplus pos (tys1,vars1) (tys2,[]) = 
    let
        val dups = List.mapPartial
                   (fn (k,t) => 
                       case List.find (fn (k',t') => k = k') tys2
                        of SOME (k',t') => SOME (k,t,t') 
                         | NONE => NONE)
                    tys1
        val mismatched = List.filter (fn (k,t,t') => t<>t') dups
        val () = if mismatched = [] then () else
                 raise Untypeable(pos,"Matching labels should have same type: " 
                                      ^ (concatWith ", " (map (fn (k,t,t')=> k ^" ("^(tytostring t)^" vs "^(tytostring t')^")")
                                                              dups)))
        val tys1' = List.filter (fn (k,_) => not (List.exists (fn (k',_,_)=>k=k') dups)) tys1
                   
		   (*        val dups = duplicates id vars1 vars2
		    val () = if dups = [] then () else
			raise Untypeable(pos,"Row variables should be distinct: " 
                                             ^ (concatWith ", " dups))
                    *)

    in (tys1'@tys2, vars1) end
  | prod_oplus pos _ (_,vars) = raise Untypeable(pos, "`Right hand' type in merge or extend contains row variables: "
						      ^ (concatWith ", " vars))

(* Recursive type unfolding *)

fun tySubst (x,t) (NamedTy y) = if y=x then t else NamedTy y
  | tySubst xt IntTy = IntTy
  | tySubst xt BoolTy = BoolTy
  | tySubst xt ForestTy = ForestTy
  | tySubst xt (ProdTy (tys,other)) = 
               ProdTy (map (fn (l,t)=>(l,tySubst xt t)) tys, other)
  | tySubst xt (SumTy tys) = SumTy (map (fn (l,t)=>(l,tySubst xt t)) tys)
  | tySubst xt (BangTy ty) = BangTy (tySubst xt ty)
  | tySubst xt (FunTy (ty1,ty2)) = FunTy (tySubst xt ty1, tySubst xt ty2)
  | tySubst xt (RecTy (x,ty)) = RecTy (x, tySubst xt ty)
  | tySubst xt (ClassImplTy (lo,ty1,ty2,ty3)) =
               ClassImplTy (lo, tySubst xt ty1, tySubst xt ty2, tySubst xt ty3)
  | tySubst xt (MergeTy (ty1,ty2)) = MergeTy (tySubst xt ty1, tySubst xt ty2)

fun recbody pos tenv (t as RecTy(x,t')) = tySubst (x,t) t'
  | recbody pos tenv (NamedTy x) = 
    (case fetchRest x tenv of
	 SOME (t,tenv') => recbody pos tenv' t
       | NONE => raise Untypeable (pos, "Unknown fold/unfold type"^x))
  | recbody pos _ _ = raise Untypeable (pos, "Fold/unfold type is not a rectype") ;


(* Match a bunch of (component,function) pairs funs against a bunch of (component,type) pairs tys - should pair off 1-1, 
 * return SOME list resulting from applying each function to the matching type, if each function successfully returns
 * SOME result. Means we can disregard ordering at point of use - returned order is that of the function list supplied.
 *)
(* Coding pattern here and below: mapPartial, and check if the resulting list has the correct length, so that we either get
 * SOME list if the function is always successful, or NONE if it is ever unsuccessful 
 *)
fun prod_match tys funs = 
   let val res = List.mapPartial (fn (k, f) => case List.find (fn (k', _) => k'=k) tys
						of SOME (_,t) => f t
						 | NONE => NONE) funs
   in if length res = length tys andalso length tys = length funs then SOME res else NONE 
   end
(* Special case where no filtering is required, just return types in tys in order of components in names.  *)
fun prod_select tys names = prod_match tys (map (fn k => (k, SOME)) names)
fun prodty_select (ProdTy(tys,_)) names = prod_select tys names 
  | prodty_select _ _ = NONE


(* Miscellaneous bits needed for class sugar *)

fun compsOf (ProdTy (complist,_)) = map fst complist
  | compsOf _ = []  (* better, raise error *)

exception MissingComponentError of label
exception WrongTypeForm of ty

val unitTy = ProdTy ([],[])
fun arrowTy s t = BangTy (FunTy (s,t))

fun RW s = ProdTy ([("read", arrowTy unitTy s), 
		    ("write", arrowTy s unitTy)], 
		   [])

fun RWall (ProdTy (complist,_)) = 
    ProdTy (map (fn (k,s) => (k,RW s)) complist, [])
  | RWall t = raise WrongTypeForm t

fun destBang (BangTy t) = t
  | destBang t = raise WrongTypeForm t  (* should handle and report *)
fun destLeft (BangTy (FunTy (t,_))) = t
  | destLeft t = raise WrongTypeForm t
fun destRight (BangTy (FunTy (_,t))) = t
  | destRight t = raise WrongTypeForm t

fun argtype (ProdTy (complist,_)) m =
    destLeft (Pervasives.errorLookup MissingComponentError complist m)
  | argtype _ m = raise MissingComponentError m

fun VS(t,s) = ProdTy ([("value",t),("state",s)], [])

fun naturalTy false (ProdTy (complist,_)) pi' =
    let val rwall = RWall pi' 
    in ProdTy (map (fn (k,t) => (k, arrowTy rwall (destBang t))) complist,
	       [])
    end
  | naturalTy true (ProdTy (complist,_)) pi' =
    let val rwall = RWall pi' 
    in ProdTy (map (fn (k,t) => (k, arrowTy (VS (destLeft t, pi'))
				            (VS (destRight t, pi'))))
		   complist,
	       [])
    end
  | naturalTy _ t _ = raise WrongTypeForm t

exception NotProdTy
exception BadMergeTy
fun expandToProdTy _ (t as ProdTy _) = t
  | expandToProdTy tenv (NamedTy v) =
    (case fetchRest v tenv of SOME (ty,tenv') => expandToProdTy tenv' ty
			    | NONE => raise NotProdTy)
  | expandToProdTy tenv (MergeTy (t1,t2)) =
    (case (expandToProdTy tenv t1, expandToProdTy tenv t2) of
	(ProdTy (l1,[]), ProdTy (l2,[])) =>
            if disjoint (map fst l1) (map fst l2)
            then ProdTy (l1@l2,[]) else raise BadMergeTy
      | _ => raise BadMergeTy)
  | expandToProdTy _ _ = raise NotProdTy

exception NotClassImplTy
fun destClassImplTy tenv (ClassImplTy (lo,tf,tm,tk)) =
    (lo, expandToProdTy tenv tf, expandToProdTy tenv tm, tk)
  | destClassImplTy tenv (NamedTy v) =
    (case fetchRest v tenv of SOME (ty,tenv') => destClassImplTy tenv' ty
			    | NONE => raise NotClassImplTy)
  | destClassImplTy _ _ = raise NotClassImplTy

fun checkExtends tenv pi pi' = true  (* dummy for now *)

fun mergeTypes tenv pi pi' =
    ProdTy (componentsOf tenv pi @ componentsOf tenv pi', [])
    (* unsafe version for now *)

end ;
